class ParserMSH:
    def __init__(self, mesh_file_path):
        """
        Read msh file and returns mesh format and physical names as class attributes.
        :param mesh_file_path: Full path to .msh file, including file name and extension.
        """
        self._mesh_format_markers = {'s': '$MeshFormat', 'e': '$EndMeshFormat'}
        self._physical_name_markers = {'s': 'PhysicalNames', 'e': '$EndPhysicalNames'}

        with open(mesh_file_path) as f:
            self._contents = f.read()

    def __get_content(self, markers_dict):
        """
        Gets text string between two markers specified in markers_dict
        """
        return self._contents[self._contents.find(markers_dict['s']) + len(markers_dict['s']):self._contents.find(markers_dict['e'])]

    @property
    def mesh_format(self):
        """
        Parse mesh_generators field and assign it to the class attribute
        """
        return self.__get_content(self._mesh_format_markers)

    @property
    def physical_names(self):
        """
        Parse physical_names field and assign it to the class attribute
        """
        return self.__get_content(self._physical_name_markers)



import os
import timeit
from pathlib import Path

import gmsh

from fiqus.utils.Utils import FilesAndFolders as uff
from fiqus.utils.Utils import GmshUtils
from fiqus.data.RegionsModelFiQuS import RegionsModel
from fiqus.pro_assemblers.ProAssembler import ASS_PRO as ass_pro


class RunGetdpCCT:
    def __init__(self, fdm, settings: dict = None, verbose=True):
        """
        Class to preparing brep files by adding terminals.
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.settings = settings
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.mesh_folder = Path(self.model_folder).parent
        self.verbose = verbose
        self.cctrm = uff.read_data_from_yaml(os.path.join(self.model_folder, f'{self.magnet_name}.regions'), RegionsModel)
        self.ap = ass_pro(os.path.join(self.model_folder, self.magnet_name))
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()
        self.pos_names = []
        # for variable, volume, file_ext in zip(self.cctdm.solve.variables, self.cctdm.solve.volumes, self.cctdm.solve.file_exts):
        #     self.pos_names.append(f'{variable}_{volume}.{file_ext}')

    def assemble_pro(self):
        if self.verbose:
            print('Assembling Pro File Started')
            start_time = timeit.default_timer()
        if self.verbose:
            print('Assembling pro file')
        self.ap.assemble_combined_pro(template=self.cctdm.solve.pro_template, rm=self.cctrm, dm=self.cctdm)
        if self.verbose:
            print(f'Assembling Pro File Took {timeit.default_timer() - start_time:.2f} s')

    def solve_and_postprocess(self, gui=False):
        command = "-solve -v2 -pos"
        model_file = 'Center_line.csv'
        self._run(command=command, model_file=model_file, gui=gui)

    def postprocess(self, gui=False):
        command = "-v2 -pos"
        model_file = 'Test.dat'
        self._run(command=command, model_file=model_file, gui=gui)

    def _run(self, command, model_file, gui=False):
        if self.verbose:
            print('Solving Started !!!')
            start_time = timeit.default_timer()
        getdp_binary = self.settings['GetDP_path']
        gmsh.onelab.run(f"{self.magnet_name}", f"{getdp_binary} {os.path.join(self.model_folder, self.magnet_name)}.pro {command} -msh {os.path.join(self.mesh_folder, self.magnet_name)}.msh")
        gmsh.onelab.setChanged("GetDP", 0)

        self.model_file = os.path.join(self.model_folder, model_file)
        if self.verbose:
            print(''
                  f'Solving Took {timeit.default_timer() - start_time:.2f} s !!!'
                  '')
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()



import os
import getpass

from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.utils.Utils import CheckForExceptions as Check
from fiqus.data import DataFiQuS as dF
from fiqus.mains.MainMultipole import MainMultipole
from fiqus.mains.MainCCT import MainCCT


class MainFiQuS:
    def __init__(self, input_file_path: str = None, model_folder: str = None, verbose: bool = True, fdm=None):
        """
        Main class for working with FiQuS simulations
        :param input_file_path: input file name
        :param verbose: if True, more info is printed in the console
        """
        self.start_folder = os.getcwd()
        self.wrk_folder = model_folder
        self.verbose = verbose
        if self.verbose:
            Util.print_welcome_graphics()

        # Load yaml input file
        if not fdm:
            self.fdm = Util.read_data_from_yaml(input_file_path, dF.FDM)
        else:
            self.fdm = fdm

        # Check for input errors
        Check.check_inputs(self.fdm.run)

        # Initialize Main object
        if self.fdm.magnet.type == 'CCT_straight':
            self.main_magnet = MainCCT(fdm=self.fdm, verbose=verbose)

        elif self.fdm.magnet.type == 'multipole':
            # Load settings
            self.sdm = Util.read_data_from_yaml(f'{input_file_path[:-5]}.set', dF.FiQuSSettings)

            self.main_magnet = MainMultipole(fdm=self.fdm, sdm=self.sdm, rgd_path=f'{input_file_path[:-5]}.geom',
                                             verbose=verbose)

        else:
            raise ValueError(f'FiQuS does not support magnet type: {self.fdm.magnet.type}!')

        # Load user paths for executables and additional files
        user_name = getpass.getuser()
        if verbose:
            print(f'FiQuS is running on machine with user name: {user_name}')
        if user_name in ['root', 'MP-WIN-02$']:
            user_name = 'SYSTEM'
        path_to_settings_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), "tests", f"settings.{user_name}.yaml")
        if verbose:
            print(f'FiQuS is using settings file: {path_to_settings_file}')
        self.main_magnet.settings = Util.read_data_from_yaml(path_to_settings_file, dict)

        # Save Model/Geometry/Mesh/Solution folder paths
        Util.prep_folder(self.wrk_folder)
        self.save_folders()

        # Build magnet
        self.summary = dict.fromkeys(['SJ', 'SICN', 'SIGE', 'Gamma', 'nodes', 'solution_time', 'overall_error', 'minimum_diff', 'maximum_diff'])
        self.build_magnet()

    def save_folders(self):
        def _check_and_generate_path(folder_type: str = None):
            if folder_type == 'Geometry':
                folder = self.wrk_folder
                ref_nr = self.fdm.run.geometry
            elif folder_type == 'Mesh':
                folder = self.main_magnet.geom_folder
                ref_nr = self.fdm.run.mesh
            elif folder_type == 'Solution':
                folder = self.main_magnet.mesh_folder
                ref_nr = self.fdm.run.solution
            else:
                raise Exception('Incompatible type.')

            required_folder = folder_type in required_folders
            if self.fdm.run.overwrite and folder_type == (required_folders[0] if required_folders else None):
                Check.check_overwrite_conditions(folder_type=folder_type, folder=folder, ref_nr=ref_nr)
            return Util.get_folder_path(folder_type=folder_type, folder=folder, ref_nr=ref_nr,
                                        overwrite=self.fdm.run.overwrite, required_folder=required_folder)

        if self.fdm.run.type == 'start_from_yaml':
            required_folders = ['Geometry', 'Mesh', 'Solution']
        elif self.fdm.run.type == 'geometry_only':
            required_folders = [] if self.fdm.run.geometry and not self.fdm.run.overwrite else ['Geometry']
        elif self.fdm.run.type == 'mesh_and_solve_with_post_process_python':
            required_folders = ['Mesh', 'Solution']
        elif self.fdm.run.type == 'mesh_only':
            required_folders = [] if self.fdm.run.mesh and not self.fdm.run.overwrite else ['Mesh']
        elif self.fdm.run.type in ['solve_with_post_process_python', 'solve_only']:
            required_folders = ['Solution']
        else:  # post_process_getdp_only or post_process_python_only or plot_python
            required_folders = []

        fdm = self.main_magnet.fdm.magnet
        self.main_magnet.geom_folder = _check_and_generate_path(folder_type='Geometry')
        Util.write_data_to_yaml(os.path.join(self.main_magnet.geom_folder, 'geometry.yaml'), fdm.geometry.dict())
        if not self.fdm.run.type == 'geometry_only':
            self.main_magnet.mesh_folder = _check_and_generate_path(folder_type='Mesh')
            Util.write_data_to_yaml(os.path.join(self.main_magnet.mesh_folder, 'mesh.yaml'), fdm.mesh.dict())
        if not (self.fdm.run.type == 'geometry_only' or self.fdm.run.type == 'mesh_only'):
            self.main_magnet.solution_folder = _check_and_generate_path(folder_type='Solution')
            Util.write_data_to_yaml(os.path.join(self.main_magnet.solution_folder, 'solve.yaml'), fdm.solve.dict())
            Util.write_data_to_yaml(os.path.join(self.main_magnet.solution_folder, 'postproc.yaml'), fdm.postproc.dict())

    def build_magnet(self):
        if self.fdm.run.type == 'start_from_yaml':  # needs 3 files (yaml, set, geom)
            self.main_magnet.generate_geometry()
            self.main_magnet.pre_process()
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
                self.summary[key] = value
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'geometry_only':
            if len(os.listdir(self.main_magnet.geom_folder)) == 1:
                self.main_magnet.generate_geometry()  # needs 3 files (yaml, set, geom)
                self.main_magnet.pre_process(gui=self.main_magnet.fdm.run.launch_gui)
            else:
                self.main_magnet.load_geometry(gui=self.main_magnet.fdm.run.launch_gui)  # needs 2 files (yaml, brep)
        elif self.fdm.run.type == 'mesh_and_solve_with_post_process_python':  # needs 5 files (yaml, strs/map2d, set, brep, aux)
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
                self.summary[key] = value
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'mesh_only':
            if len(os.listdir(self.main_magnet.mesh_folder)) == 1:
                self.main_magnet.load_geometry()  # needs 3 files (yaml, brep, aux)
                for key, value in self.main_magnet.mesh(gui=self.main_magnet.fdm.run.launch_gui).items():
                    self.summary[key] = value
            else:
                self.main_magnet.load_mesh(gui=self.main_magnet.fdm.run.launch_gui)  # needs 2 files (yaml, msh)
        elif self.fdm.run.type == 'solve_with_post_process_python':  # needs 5 files (yaml, strs/map2d, set, msh, reg)
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'solve_only':  # needs 5 files (yaml, strs/map2d, set, msh, reg)
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == 'post_process_getdp_only':
            self.main_magnet.post_process_getdp(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == 'post_process_python_only':
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'plot_python':
            self.main_magnet.plot_python()
        os.chdir(self.start_folder)

if __name__ == "__main__":
    # if len(sys.argv) < 4:
    #     mp = MainFiQuS(sys.argv[1], sys.argv[2])

    magnet = 'MQXA'
    yaml_file = f'{magnet}.yaml'
    output_folder = os.getcwd()
    sim_result = MainFiQuS(input_file_path=yaml_file, model_folder=os.path.join(output_folder, magnet))



import math
import os
import timeit
import json
import gmsh
import numpy as np
from fiqus.utils.Utils import FilesAndFolders as uff
from fiqus.utils.Utils import GmshUtils
from fiqus.data.DataWindingsCCT import WindingsInformation         # for volume information


class Pre_Process:
    def __init__(self, fdm, verbose=True):
        """
        Class to preparing brep files by adding terminals.
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        winding_info_file = os.path.join(self.model_folder, f'{self.magnet_name}.wi')
        self.cctwi = uff.read_data_from_yaml(winding_info_file, WindingsInformation)
        self.verbose = verbose
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()

    def calculate_normals(self, gui=False):
        """
        Calculates normals for the cct channel directions, i.e. along winding direction, along radial direction of the former (height) and along axial direction (width). Normals are saved to a .json
        file and used later for post-processing of magnetic field into components along channel length, height and width. Note that this function does not give the correct 'sign of normals', i.e.
        normals facing inwards or outwards of the surface are not properly distinguished. The normals along the length are not reliable and only along the height and width are used for calculations and
        the field along the length is taken as a remaining field.
        This function needs full geom_generators .brep file and volume information files (.vi) for each individual powered volume.
        :param gui: if True, the gmsh graphical user interface is shown at the end and normals are displayed as a view
        :return: Nothing, a file for each powered geom_generators brep is
        """
        if self.verbose:
            print('Calculating Normals Started')
            start_time = timeit.default_timer()
        gmsh.open(os.path.join(self.model_folder, f'{self.magnet_name}.brep'))

        def _calc_normals_dir(tags_for_normals_in, surfs_idx, surfs_scale):
            v_to_suf = [[0, 1, 2, 3], [0, 1, 5, 4], [4, 5, 6, 7], [3, 7, 6, 2], [0, 3, 7, 4], [1, 5, 6, 2]]
            norm_e_x = []   # normal along x of volume
            norm_e_y = []
            norm_e_z = []
            norm_dict = {}
            normals_view = []  # this remains an empty list if view is False
            coor_e_x = []   # coordinate x of the center of volume
            coor_e_y = []
            coor_e_z = []
            for vol_tag in tags_for_normals_in:
                all_surf_tags = gmsh.model.getAdjacencies(3, vol_tag)[1]
                surf_tags = [all_surf_tags[index] for index in surfs_idx]
                norm = []
                node_coord = []
                vol_line_tags = []
                for surf_tag in all_surf_tags:
                    line_tags_new = gmsh.model.getAdjacencies(2, surf_tag)[1]
                    for line_tag in line_tags_new:
                        if line_tag not in vol_line_tags:
                            vol_line_tags.append(line_tag)
                point_tags = []
                for line_tag in vol_line_tags:
                    point_tags_new = gmsh.model.getAdjacencies(1, line_tag)[1]
                    for point_tag in point_tags_new:
                        if point_tag not in point_tags:
                            point_tags.append(int(point_tag))
                for surf_i, surf_tag, scale in zip(surfs_idx, surf_tags, surfs_scale):
                    p_idx = v_to_suf[surf_i]
                    s_node_coord = []
                    for p_i in p_idx:
                        xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, point_tags[p_i])
                        s_node_coord.append((xmin + xmax) / 2)
                        s_node_coord.append((ymin + ymax) / 2)
                        s_node_coord.append((zmin + zmax) / 2)
                    parametricCoord = gmsh.model.getParametrization(2, surf_tag, s_node_coord)
                    s_norm = gmsh.model.getNormal(surf_tag, parametricCoord)
                    norm.extend(scale*s_norm)
                    node_coord.extend(s_node_coord)
                coor_s_x = []   # coordinates surface x
                coor_s_y = []
                coor_s_z = []
                norm_s_x = []   # normals surface x
                norm_s_y = []
                norm_s_z = []

                for i in range(0, len(node_coord), 3):
                    coor_s_x.append(node_coord[i])
                    coor_s_y.append(node_coord[i+1])
                    coor_s_z.append(node_coord[i+2])
                    norm_s_x.append(norm[i])
                    norm_s_y.append(norm[i+1])
                    norm_s_z.append(norm[i+2])

                coor_e_x.append(np.mean(coor_s_x))
                coor_e_y.append(np.mean(coor_s_y))
                coor_e_z.append(np.mean(coor_s_z))
                # norm_e_x.append(np.mean(norm_s_x))
                # norm_e_y.append(np.mean(norm_s_y))
                # norm_e_z.append(np.mean(norm_s_z))

                # norm_e_x.append(np.sqrt(np.sum(np.square(norm_s_x)))/(2*np.sqrt(2)))
                # norm_e_y.append(np.sqrt(np.sum(np.square(norm_s_y)))/(2*np.sqrt(2)))
                # norm_e_z.append(np.sqrt(np.sum(np.square(norm_s_z)))/(2*np.sqrt(2)))
                v_x = np.sum(norm_s_x)
                v_y = np.sum(norm_s_y)
                v_z = np.sum(norm_s_z)
                ampl = math.sqrt(v_x**2 + v_y**2 + v_z**2)

                norm_e_x.append(v_x/ampl)
                norm_e_y.append(v_y/ampl)
                norm_e_z.append(v_z/ampl)

            for i in range(len(coor_e_x)):
                normals_view.append(coor_e_x[i])
                normals_view.append(coor_e_y[i])
                normals_view.append(coor_e_z[i])
                normals_view.append(norm_e_x[i])
                normals_view.append(norm_e_y[i])
                normals_view.append(norm_e_z[i])
            norm_dict['x'] = coor_e_x
            norm_dict['y'] = coor_e_y
            norm_dict['z'] = coor_e_z
            norm_dict['n_x'] = norm_e_x
            norm_dict['n_y'] = norm_e_y
            norm_dict['n_z'] = norm_e_z
            return norm_dict, normals_view
        """
        This is helper function called in a loop below.
        """
        max_tag = 0
        for f_name in self.cctwi.w_names+self.cctwi.f_names:
            vol_tags = json.load(open(os.path.join(self.model_folder, f'{f_name}.vi')))
            export_tags = vol_tags['export']
            tags_for_normals = [e + max_tag for e in export_tags]
            max_tag = np.max(vol_tags['all']) + max_tag
            surfs_idx_l = [0, 2]    # along length of the former groove
            if f_name in self.cctwi.w_names:
                surfs_scale_l = [1, 1]
            elif f_name in self.cctwi.f_names:
                surfs_scale_l = [1, -1]     # change direction for fqpl
            norm_l, norm_view_l = _calc_normals_dir(tags_for_normals, surfs_idx_l, surfs_scale_l)
            surfs_idx_h = [1, 3]         # along height of the former groove
            surfs_scale_h = [1, -1]
            norm_h, norm_view_h = _calc_normals_dir(tags_for_normals, surfs_idx_h, surfs_scale_h)
            surfs_idx_w = [4, 5]        # along width of the former groove
            surfs_scale_w = [1, -1]
            norm_w, norm_view_w = _calc_normals_dir(tags_for_normals, surfs_idx_w, surfs_scale_w)
            normals_dict = {'normals_l': norm_l, 'normals_h': norm_h, 'normals_w': norm_w}
            json.dump(normals_dict, open(f"{os.path.join(self.model_folder, f_name)}.normals", 'w'))
            if gui:
                normals_all = [norm_view_l, norm_view_h, norm_view_w]
                self.__add_normals_view(f_name, normals_all)
        if self.verbose:
            print(f'Calculating Normals Took {timeit.default_timer() - start_time:.2f} s')
        if gui:
            self.gu.launch_interactive_GUI()

    @staticmethod
    def __add_normals_view(name, normals_all, norm_list=[0, 1, 2]):
        """
        THis adds new view in gmsh.
        :param name: name of view
        :param normals_all: dictionary with normals
        :param norm_list: which normals to plot. Default is: [0, 1, 2] corresponds to [n_l, n_h, n_w]. If this array is shorter the corresponding normals are skipped in views.
        :return:
        """
        norm_names_all = [f"{name}_n_l", f"{name}_n_h", f"{name}_n_w"]
        norm_names = [norm_names_all[index] for index in norm_list]
        normals = [normals_all[index] for index in norm_list]
        for view_name, view_data in zip(norm_names, normals):
            gmsh.view.addListData(gmsh.view.add(view_name), "VP", len(view_data) // 6, view_data)
        gmsh.model.occ.synchronize()



import os
import gmsh
from numpy import square as square
from numpy import sqrt as sqrt
import json

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.utils.Utils import GeometricFunctions as Func
from fiqus.data import DataFiQuS as dF
from fiqus.data import DataMultipole as dM


class Geometry:
    def __init__(self, data: dF.FDM() = None, geom: dF.FiQuSGeometry() = None, sett: dF.FiQuSSettings() = None,
                 geom_folder: str = None, verbose: bool = False):
        """
        Class to generate geometry
        :param data: FiQuS data model
        :param geom: ROXIE geometry data
        :param sett: settings data model
        :param verbose: If True more information is printed in python console.
        """
        self.data: dF.FDM() = data
        self.geom: dF.FiQuSGeometry() = geom
        self.sett: dF.FiQuSSettings() = sett
        self.geom_folder = geom_folder
        self.verbose: bool = verbose

        self.md = dM.MultipoleData()

        self.gu = GmshUtils(self.geom_folder, self.verbose)
        self.gu.initialize()
        self.occ = gmsh.model.occ

        self.model_file = f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.brep"
        self.iL, self.iR, self.oL, self.oR, self.iLr, self.iRr, self.oLr, self.oRr = [], [], [], [], [], [], [], []

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def saveHalfTurnCornerPositions(self):
        json.dump({'iL': self.iL, 'iR': self.iR, 'oL': self.oL, 'oR': self.oR,
                   'iLr': self.iLr, 'iRr': self.iRr, 'oLr': self.oLr, 'oRr': self.oRr},
                  open(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.crns", 'w'))

    def saveStrandPositions(self):
        ht_nr = 0
        parser_x, parser_y, blocks, ht = [], [], [], []
        for eo in self.geom.Roxie_Data.coil.electrical_order:
            block = self.geom.Roxie_Data.coil.coils[eo.coil].poles[eo.pole].layers[eo.layer].windings[
                eo.winding].blocks[eo.block]
            for halfTurn_nr, halfTurn in block.half_turns.items():
                ht_nr += 1
                for strand_group_nr, strand_group in halfTurn.strand_groups.items():
                    for strand_nr, strand in strand_group.strand_positions.items():
                        blocks.append(eo.block)
                        ht.append(ht_nr)
                        parser_x.append(strand.x)
                        parser_y.append(strand.y)
        json.dump({'x': parser_x, 'y': parser_y, 'block': blocks, 'ht': ht},
                  open(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.strs", 'w'))

    def saveBoundaryRepresentationFile(self):
        self.occ.synchronize()
        gmsh.write(self.model_file)
        gmsh.clear()

    def loadBoundaryRepresentationFile(self):
        gmsh.option.setString('Geometry.OCCTargetUnit', 'M')  # set units to meters
        gmsh.open(self.model_file)

    def saveAuxiliaryFile(self):
        Util.write_data_to_yaml(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.aux", self.md.dict())

    def constructIronGeometry(self):
        """
            Generates points, hyper lines, and curve loops for the iron yoke
        """
        iron = self.geom.Roxie_Data.iron
        self.md.geometries.iron.quadrants = {1: dM.Region(), 2: dM.Region(), 3: dM.Region(), 4: dM.Region()}
        quadrants = self.md.geometries.iron.quadrants

        lc = 1e-2
        for point_name, point in iron.key_points.items():
            quadrants[1].points[point_name] = self.occ.addPoint(point.x, point.y, 0, lc)

            if point.x == 0.:
                quadrants[2].points[point_name] = quadrants[1].points[point_name]
            else:
                quadrants[2].points[point_name] = self.occ.copy([(0, quadrants[1].points[point_name])])[0][1]
                self.occ.mirror([(0, quadrants[2].points[point_name])], 1, 0, 0, 0)
            if point.y == 0.:
                quadrants[3].points[point_name] = quadrants[2].points[point_name]
                quadrants[4].points[point_name] = quadrants[1].points[point_name]
            else:
                quadrants[3].points[point_name] = self.occ.copy([(0, quadrants[2].points[point_name])])[0][1]
                self.occ.mirror([(0, quadrants[3].points[point_name])], 0, 1, 0, 0)
                quadrants[4].points[point_name] = self.occ.copy([(0, quadrants[1].points[point_name])])[0][1]
                self.occ.mirror([(0, quadrants[4].points[point_name])], 0, 1, 0, 0)

        mirror_x = [1, -1, -1, 1]
        mirror_y = [1, 1, -1, -1]
        for line_name, line in iron.hyper_lines.items():
            if line.type == 'line':
                for quadrant, qq in quadrants.items():
                    if quadrant == 1:
                        qq.lines[line_name] = self.occ.addLine(qq.points[line.kp1], qq.points[line.kp2])
                    elif quadrant == 2:
                        if iron.key_points[line.kp1].x == 0. and iron.key_points[line.kp2].x == 0.:
                            qq.lines[line_name] = quadrants[1].lines[line_name]
                        else:
                            qq.lines[line_name] = self.occ.addLine(qq.points[line.kp1], qq.points[line.kp2])
                    elif quadrant == 3:
                        if iron.key_points[line.kp1].y == 0. and iron.key_points[line.kp2].y == 0.:
                            qq.lines[line_name] = quadrants[2].lines[line_name]
                        else:
                            qq.lines[line_name] = self.occ.addLine(qq.points[line.kp1], qq.points[line.kp2])
                    else:
                        if iron.key_points[line.kp1].y == 0. and iron.key_points[line.kp2].y == 0.:
                            qq.lines[line_name] = quadrants[1].lines[line_name]
                        else:
                            qq.lines[line_name] = self.occ.addLine(qq.points[line.kp1], qq.points[line.kp2])

            elif line.type == 'arc':
                center = Func.arc_center_from_3_points(
                    iron.key_points[line.kp1], iron.key_points[line.kp3], iron.key_points[line.kp2])
                new_point_name = 'kp' + line_name + '_center'
                quadrants[1].points[new_point_name] = self.occ.addPoint(center[0], center[1], 0)
                # gmsh.model.setEntityName(0, gm.iron.quadrants[1].points[new_point_name], 'iron_' + new_point_name)
                if center[0] == 0.:
                    quadrants[2].points[new_point_name] = quadrants[1].points[new_point_name]
                else:
                    quadrants[2].points[new_point_name] = self.occ.copy([(0, quadrants[1].points[new_point_name])])[0][1]
                    self.occ.mirror([(0, quadrants[2].points[new_point_name])], 1, 0, 0, 0)
                if center[1] == 0.:
                    quadrants[3].points[new_point_name] = quadrants[2].points[new_point_name]
                    quadrants[4].points[new_point_name] = quadrants[1].points[new_point_name]
                else:
                    quadrants[3].points[new_point_name] = self.occ.copy([(0, quadrants[2].points[new_point_name])])[0][1]
                    self.occ.mirror([(0, quadrants[3].points[new_point_name])], 0, 1, 0, 0)
                    quadrants[4].points[new_point_name] = self.occ.copy([(0, quadrants[1].points[new_point_name])])[0][1]
                    self.occ.mirror([(0, quadrants[4].points[new_point_name])], 0, 1, 0, 0)

                for quadrant, qq in quadrants.items():
                    qq.lines[line_name] = self.occ.addCircleArc(
                        qq.points[line.kp1], qq.points[new_point_name], qq.points[line.kp2])

            elif line.type == 'circle':
                pt1 = iron.key_points[line.kp1]
                pt2 = iron.key_points[line.kp2]
                center = [(pt1.x + pt2.x) / 2, (pt1.y + pt2.y) / 2]
                radius = (sqrt(square(pt1.x - center[0]) + square(pt1.y - center[1])) +
                          sqrt(square(pt2.x - center[0]) + square(pt2.y - center[1]))) / 2

                for quadrant, qq in quadrants.items():
                    qq.lines[line_name] = self.occ.addCircle(
                        mirror_x[quadrant - 1] * center[0], mirror_y[quadrant - 1] * center[1], 0, radius)
                    qq.points['kp' + line_name] = len(qq.points) + 1

            else:
                raise ValueError('Hyper line {} not supported'.format(line.type))

        for quadrant, qq in quadrants.items():
            for area_name, area in iron.hyper_areas.items():
                qq.areas[area_name] = dM.Area(loop=self.occ.addCurveLoop([qq.lines[line] for line in area.lines]))
                if (iron.hyper_areas[area_name].material not in self.md.domains.groups_surfaces and
                        iron.hyper_areas[area_name].material != 'BH_air'):
                    self.md.domains.groups_surfaces[iron.hyper_areas[area_name].material] = []

    def constructWedgeGeometry(self):
        """
            Generates points, hyper lines, and curve loops for the wedges
        """
        wedges = self.geom.Roxie_Data.wedges
        for i in wedges:
            kp0_inner = self.occ.addPoint(wedges[i].corrected_center.inner.x, wedges[i].corrected_center.inner.y, 0)
            kp0_outer = self.occ.addPoint(wedges[i].corrected_center.outer.x, wedges[i].corrected_center.outer.y, 0)
            arg = [self.occ.addPoint(wedges[i].corners.iL.x, wedges[i].corners.iL.y, 0),
                   self.occ.addPoint(wedges[i].corners.iR.x, wedges[i].corners.iR.y, 0),
                   self.occ.addPoint(wedges[i].corners.oL.x, wedges[i].corners.oL.y, 0),
                   self.occ.addPoint(wedges[i].corners.oR.x, wedges[i].corners.oR.y, 0)]

            left = self.occ.addLine(arg[0], arg[2])
            right = self.occ.addLine(arg[1], arg[3])
            inner = self.occ.addCircleArc(arg[0], kp0_inner, arg[1])
            outer = self.occ.addCircleArc(arg[2], kp0_outer, arg[3])
            self.md.geometries.wedges.areas[i] = dM.Area(loop=self.occ.addCurveLoop([inner, right, outer, left]))

    def constructCoilGeometry(self):
        """
            Generates points, hyper lines, and curve loops for the coil half-turns
        """
        frac = 1  # self.frac if self.frac else 1
        for coil_nr, coil in self.geom.Roxie_Data.coil.coils.items():
            self.md.geometries.coil.coils[coil_nr] = dM.Pole()
            coils = self.md.geometries.coil.coils[coil_nr]
            for pole_nr, pole in coil.poles.items():
                coils.poles[pole_nr] = dM.Layer()
                poles = coils.poles[pole_nr]
                for layer_nr, layer in pole.layers.items():
                    poles.layers[layer_nr] = dM.Winding()
                    layers = poles.layers[layer_nr]
                    for winding_nr, winding in layer.windings.items():
                        layers.windings[winding_nr] = dM.Block(conductor_name=winding.conductor_name,
                                                               conductors_number=winding.conductors_number)
                        windings = layers.windings[winding_nr]
                        ll = self.sett.Model_Data_GS.conductors[winding.conductor_name].cable.bare_cable_height_mean
                        for block_key, block in winding.blocks.items():
                            windings.blocks[block_key] = dM.BlockData(current_sign=block.current_sign)
                            hts = windings.blocks[block_key].half_turns

                            for halfTurn_nr, halfTurn in block.half_turns.items():
                                ht = str(halfTurn_nr)
                                hts.areas[ht] = dM.Area()
                                hf_current = halfTurn.corners.insulated
                                if halfTurn_nr == 1:
                                    bc = block.block_corners
                                    if (hf_current.iL.y < 0 and block_key == max(winding.blocks.keys())) or \
                                            (hf_current.iL.y > 0 and block_key == min(winding.blocks.keys())):
                                        hts.points[ht + 'i'] = self.occ.addPoint(bc.iR.x, bc.iR.y, 0, ll / frac)
                                        hts.points[ht + 'o'] = self.occ.addPoint(bc.oR.x, bc.oR.y, 0, ll / frac)
                                    else:
                                        hts.points[ht + 'i'] = self.occ.addPoint(bc.iL.x, bc.iL.y, 0, ll / frac)
                                        hts.points[ht + 'o'] = self.occ.addPoint(bc.oL.x, bc.oL.y, 0, ll / frac)
                                    hts.lines[ht + 'r'] = self.occ.addLine(hts.points[ht + 'i'], hts.points[ht + 'o'])

                                if halfTurn_nr == winding.conductors_number:
                                    bc = block.block_corners

                                    if (hf_current.iL.y < 0 and block_key == max(winding.blocks.keys())) or \
                                            (hf_current.iL.y > 0 and block_key == min(winding.blocks.keys())):
                                        hts.points['end_i'] = self.occ.addPoint(bc.iL.x, bc.iL.y, 0, ll/frac)
                                        hts.points['end_o'] = self.occ.addPoint(bc.oL.x, bc.oL.y, 0, ll/frac)
                                    else:
                                        hts.points['end_i'] = self.occ.addPoint(bc.iR.x, bc.iR.y, 0, ll/frac)
                                        hts.points['end_o'] = self.occ.addPoint(bc.oR.x, bc.oR.y, 0, ll/frac)

                                    hts.lines[ht + 'i'] = self.occ.addLine(hts.points['end_i'], hts.points[ht + 'i'])
                                    hts.lines[ht + 'o'] = self.occ.addLine(hts.points['end_o'], hts.points[ht + 'o'])
                                    hts.lines['end'] = self.occ.addLine(hts.points['end_i'], hts.points['end_o'])

                                    # For plotting only
                                    if pole_nr == 1:
                                        self.occ.synchronize()
                                        self.iL.append(list(gmsh.model.getValue(0, hts.points['end_i'], [])[:-1]))
                                        self.oL.append(list(gmsh.model.getValue(0, hts.points['end_o'], [])[:-1]))
                                        # if block_key == 1 or block_key == 3 or block_key == 5 or block_key == 6:
                                        #     radius = np.sqrt(np.square(iL[0] - coil.bore_center.x) +
                                        #                      np.square(iL[1] - coil.bore_center.y))
                                        #     self.ax.add_patch(
                                        #         patches.Circle((coil.bore_center.x, coil.bore_center.y), radius=radius,
                                        #                        color='b', fill=False))
                                        #     radius = np.sqrt(np.square(oL[0] - coil.bore_center.x) +
                                        #                      np.square(oL[1] - coil.bore_center.y))
                                        #     self.ax.add_patch(
                                        #         patches.Circle((coil.bore_center.x, coil.bore_center.y), radius=radius,
                                        #                        color='c', fill=False))
                                    ################

                                    left = hts.lines['end']
                                else:
                                    hf_next = block.half_turns[halfTurn_nr + 1].corners.insulated
                                    next_ht = str(halfTurn_nr + 1)
                                    mid_point_i = [(hf_next.iR.x + hf_current.iL.x) / 2,
                                                   (hf_next.iR.y + hf_current.iL.y) / 2]
                                    mid_point_o = [(hf_next.oR.x + hf_current.oL.x) / 2,
                                                   (hf_next.oR.y + hf_current.oL.y) / 2]
                                    hts.points[next_ht + 'i'] = self.occ.addPoint(mid_point_i[0], mid_point_i[1], 0,
                                                                                  ll / frac)
                                    hts.points[next_ht + 'o'] = self.occ.addPoint(mid_point_o[0], mid_point_o[1], 0,
                                                                                  ll / frac)

                                    hts.lines[ht + 'i'] = \
                                        self.occ.addLine(hts.points[next_ht + 'i'], hts.points[ht + 'i'])
                                    hts.lines[ht + 'o'] = \
                                        self.occ.addLine(hts.points[next_ht + 'o'], hts.points[ht + 'o'])
                                    hts.lines[next_ht + 'r'] = \
                                        self.occ.addLine(hts.points[next_ht + 'i'], hts.points[next_ht + 'o'])
                                    left = hts.lines[next_ht + 'r']

                                    # For plotting only
                                    if pole_nr == 1:
                                        self.occ.synchronize()
                                        self.iL.append(list(gmsh.model.getValue(0, hts.points[next_ht + 'i'], [])[:-1]))
                                        self.oL.append(list(gmsh.model.getValue(0, hts.points[next_ht + 'o'], [])[:-1]))
                                if pole_nr == 1:
                                    self.iR.append(list(gmsh.model.getValue(0, hts.points[ht + 'i'], [])[:-1]))
                                    self.oR.append(list(gmsh.model.getValue(0, hts.points[ht + 'o'], [])[:-1]))
                                    hti = halfTurn.corners.insulated
                                    self.iLr.append([hti.iL.x, hti.iL.y])
                                    self.iRr.append([hti.iR.x, hti.iR.y])
                                    self.oLr.append([hti.oL.x, hti.oL.y])
                                    self.oRr.append([hti.oR.x, hti.oR.y])
                                ################

                                hts.areas[ht].loop = self.occ.addCurveLoop([hts.lines[ht + 'i'],  # inner
                                                                            hts.lines[ht + 'r'],  # right
                                                                            hts.lines[ht + 'o'],  # outer
                                                                            left])  # left

        self.saveHalfTurnCornerPositions()

    def buildDomains(self):
        """
            Generates plane surfaces from the curve loops
        """
        iron = self.geom.Roxie_Data.iron
        gm = self.md.geometries

        for i in iron.key_points:
            gm.iron.max_radius = max(gm.iron.max_radius, max(iron.key_points[i].x, iron.key_points[i].y))

        # Create and build air far field
        radius_in = gm.iron.max_radius * (2.5 if self.data.magnet.geometry.with_iron_yoke else 6)
        gm.air_inf.lines['inner'] = self.occ.addCircle(0., 0., 0., radius_in)
        gm.air_inf.areas['inner'] = dM.Area(loop=self.occ.addCurveLoop([gm.air_inf.lines['inner']]))
        gm.air_inf.areas['inner'].surface = self.occ.addPlaneSurface([self.md.geometries.air_inf.areas['inner'].loop])
        radius_out = gm.iron.max_radius * (3.2 if self.data.magnet.geometry.with_iron_yoke else 8)
        gm.air_inf.lines['outer'] = self.occ.addCircle(0., 0., 0., radius_out)
        gm.air_inf.areas['outer'] = dM.Area(loop=self.occ.addCurveLoop([gm.air_inf.lines['outer']]))
        gm.air_inf.areas['outer'].surface = self.occ.addPlaneSurface([gm.air_inf.areas['outer'].loop,
                                                                      gm.air_inf.areas['inner'].loop])
        self.md.domains.groups_surfaces['air_inf'] = [gm.air_inf.areas['outer'].surface]

        # Build iron yoke domains
        if self.data.magnet.geometry.with_iron_yoke:
            for quadrant, qq in gm.iron.quadrants.items():
                for area_name, area in qq.areas.items():
                    build = True
                    loops = [area.loop]
                    for hole_key, hole in iron.hyper_holes.items():
                        if area_name == hole.areas[1]:
                            loops.append(qq.areas[hole.areas[0]].loop)
                        elif area_name == hole.areas[0]:  # or iron.hyper_areas[area_name].material == 'BH_air':
                            build = False
                    if build:
                        area.surface = self.occ.addPlaneSurface(loops)
                        self.md.domains.groups_surfaces[iron.hyper_areas[area_name].material].append(area.surface)

        # Build coil domains
        for coil_nr, coil in gm.coil.coils.items():
            for pole_nr, pole in coil.poles.items():
                for layer_nr, layer in pole.layers.items():
                    for winding_nr, winding in layer.windings.items():
                        # cable = self.set.Model_Data_GS.conductors[winding.conductor_name].cable
                        # ht_area = cable.bare_cable_height_mean * cable.bare_cable_width
                        for block_key, block in winding.blocks.items():
                            current_sign = '_neg' if block.current_sign < 0 else '_pos'
                            for area_name, area in block.half_turns.areas.items():
                                ht_name = 'block' + str(block_key) + '_ht' + str(area_name) + current_sign
                                # self.conductor_areas[ht_name] = ht_area
                                area.surface = self.occ.addPlaneSurface([area.loop])
                                self.md.domains.groups_surfaces[ht_name] = [area.surface]

        # Build wedges domains
        for wedge_nr, wedge in gm.wedges.areas.items():
            wedge.surface = self.occ.addPlaneSurface([wedge.loop])
            self.md.domains.groups_surfaces['wedge_' + str(wedge_nr)] = [wedge.surface]

    def updateTags(self):
        self.md.geometries.coil.electrical_order = self.geom.Roxie_Data.coil.electrical_order
        for coil_nr, coil in self.geom.Roxie_Data.coil.coils.items():
            self.md.geometries.coil.coils[coil_nr].bore_center = coil.bore_center

        for coil_nr, coil in self.md.geometries.coil.coils.items():
            for pole_nr, pole in coil.poles.items():
                for layer_nr, layer in pole.layers.items():
                    for winding_nr, winding in layer.windings.items():
                        for block_key, block in winding.blocks.items():
                            hts = block.half_turns
                            for ht_nr, ht in hts.areas.items():
                                lines_tags = gmsh.model.getAdjacencies(2, ht.surface)[1]
                                first_tag = int(min(lines_tags))
                                hts.lines[ht_nr + ('i' if ht_nr == '1' else 'r')] = first_tag
                                hts.lines[ht_nr + ('r' if ht_nr == '1' else 'i')] = first_tag + 1
                                hts.lines[ht_nr + 'o'] = first_tag + 2
                                if int(ht_nr) == len(hts.areas):
                                    hts.lines['end'] = first_tag + 3

        lines_tags = gmsh.model.getAdjacencies(2, self.md.geometries.air_inf.areas['outer'].surface)[1]
        self.md.geometries.air_inf.lines['outer'] = int(lines_tags[0])
        self.md.geometries.air_inf.lines['inner'] = int(lines_tags[1])



import os

class PlotPythonMultipole:
    def __init__(self, fdm, verbose=True):
        """
        Class for making python plots of multipole magnets
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.fdm = fdm
        self.verbose = verbose

    def dummy_plot_func(self):
        """
        not implemented yet
        :return:
        :rtype:
        """
        pass


import os
from operator import le, ge

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from fiqus.geom_generators.GeometryCCT import Winding, FQPL


class PlotPythonCCT:
    def __init__(self, fdm, verbose=True):
        """
        Class to cct models postprocessing
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name

        self.verbose = verbose
        # self.gu = GmshUtils(self.model_folder, self.verbose)
        # self.gu.initialize()

    def plot_elements_file(self, lfes=None):
        field_map_3D_csv = 'field_map_3D.csv'
        df = pd.read_csv(field_map_3D_csv, delimiter=',', engine='python')
        sAvePositions = df['sAvePositions [m]'].to_numpy(dtype='float')[:-1]  # [:-1] to remove last nan
        ph_order = df['nodeToPhysicalTurn [-]'].to_numpy(dtype='int')[:-1]  # physical turns, [:-1] to remove last nan
        el_order = df['nodeToHalfTurn [-]'].to_numpy(dtype='int')[:-1]  # electric order turns, [:-1] to remove last nan
        x3Dsurf_1 = df['x3Dsurf 1 [m]'].to_numpy(dtype='float')
        y3Dsurf_1 = df['y3Dsurf 1 [m]'].to_numpy(dtype='float')
        z3Dsurf_1 = df['z3Dsurf 1 [m]'].to_numpy(dtype='float')
        el_order_unique = np.unique(el_order)

        per_turn = False

        fig = plt.figure(figsize=(16, 10))
        ax = fig.add_subplot(111, projection='3d', proj_type='ortho')

        if per_turn:
            colors = cm.rainbow(np.linspace(0, 1, np.shape(el_order_unique)[0]))
            for idx_turn, el_u in enumerate(el_order_unique.tolist()):
                indexes = np.where(el_u == el_order)[0]
                min_idx=np.min(indexes)
                max_idx=np.max(indexes+2)
                ax.plot(x3Dsurf_1[min_idx:max_idx], y3Dsurf_1[min_idx:max_idx], z3Dsurf_1[min_idx:max_idx], color=colors[idx_turn])
        else:
            ax.plot(x3Dsurf_1, y3Dsurf_1, z3Dsurf_1)
        # colors = cm.rainbow(np.linspace(0, 1, np.shape(x3Dsurf_1)[0]))
        # # ax.plot(x3Dsurf_1, y3Dsurf_1, z3Dsurf_1)

        # for idx in range(np.shape(x3Dsurf_1)[0]-1):
        #     ax.plot([x3Dsurf_1[idx], x3Dsurf_1[idx+1]], [y3Dsurf_1[idx], y3Dsurf_1[idx+1]], [z3Dsurf_1[idx], z3Dsurf_1[idx+1]], color=colors[idx])
        # xmin, xmax, ymin, ymax, zmin, zmax = (0, 0, 0, 0, 0, 0)
        # #colors = cm.rainbow(np.linspace(0, 1, len(objects_list)))
        # ax.set_box_aspect((xmax - xmin, ymax - ymin, zmax - zmin))  # set axis to data aspect ratio
        # ax.set_xlim([xmin, xmax])
        # ax.set_ylim([ymin, ymax])
        # ax.set_zlim([zmin, zmax])
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.set_zlabel('z (m)')
        ax.view_init(elev=90, azim=0, vertical_axis='x')
        plt.show()
        pass

    def plot_elements_computed(self, lfes=None):
        """
        Makes a plot oh hexahedra by calculating 'on the fly' what is defined in the yaml data model for cct magnet.
        :param lfes: list with numbers of elements to put text labels on. For two windings an example would be [[3], [9]] to plot 3rd hex in the first winding and 9th in the second
        :return: Nothing. A matplotlib graph pops up on the screen.
        """

        windings = []
        fqpls = []
        for ww, _ in enumerate(self.cctdm.geometry.windings.names):
            winding_obj = Winding(self.cctdm, ww, post_proc=True)
            windings.append(winding_obj)
        for ff, _ in enumerate(self.cctdm.geometry.fqpls.names):
            fqpl_obj = FQPL(self.cctdm, ff)
            fqpls.append(fqpl_obj)
        objects_list = windings + fqpls
        if not lfes:
            lfes = [[] for _ in objects_list]  # make it work if lfes is not supplied as input.

        def val_check(var, array, oper, func):
            new = func(array)
            if oper(new, var):
                return new
            else:
                return var

        fig = plt.figure(figsize=(16, 10))
        ax = fig.add_subplot(111, projection='3d', proj_type='ortho')
        xmin, xmax, ymin, ymax, zmin, zmax = (0, 0, 0, 0, 0, 0)
        colors = cm.rainbow(np.linspace(0, 1, len(objects_list)))
        for obj, lfe, fc in zip(objects_list, lfes, colors):
            xs = []
            ys = []
            zs = []
            hexes = obj.hexes
            vts = obj.vertices_to_surf
            for elem_num, points_dict in hexes.items():
                list_of_coor_tuples = []
                points_dict.pop('ct', None)  # removing turn from dict
                for p_num, p_coords in points_dict.items():

                    list_of_coor_tuples.append((p_coords['x'], p_coords['y'], p_coords['z']))
                    if elem_num in lfe:
                        ax.scatter(p_coords['x'], p_coords['y'], p_coords['z'])
                        ax.text(p_coords['x'], p_coords['y'], p_coords['z'], f'{p_num}', zdir='z')
                    for coor_acc, coor_key in zip([xs, ys, zs], ['x', 'y', 'z']):
                        coor_acc.append(p_coords[coor_key])
                poly3d = [[list_of_coor_tuples[vts[ix][iy]] for iy in range(len(vts[0]))] for ix in range(len(vts))]
                ax.add_collection3d(Poly3DCollection(poly3d, edgecolors='k', facecolors=fc, linewidths=0.5, alpha=0.5))
            xmin = val_check(xmin, xs, le, np.min)
            xmax = val_check(xmax, xs, ge, np.max)
            ymin = val_check(ymin, ys, le, np.min)
            ymax = val_check(ymax, ys, ge, np.max)
            zmin = val_check(zmin, zs, le, np.min)
            zmax = val_check(zmax, zs, ge, np.max)
        ax.set_box_aspect((xmax - xmin, ymax - ymin, zmax - zmin))  # set axis to data aspect ratio
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.set_zlim([zmin, zmax])
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.set_zlabel('z (m)')
        ax.view_init(elev=90, azim=0, vertical_axis='x')
        plt.show()


import math
import os
import timeit
import json
from typing import List, Any
from pathlib import Path

import gmsh
import numpy as np
from fiqus.utils.Utils import FilesAndFolders as uff
from fiqus.utils.Utils import GmshUtils
from fiqus.data.DataWindingsCCT import WindingsInformation  # for volume information
from fiqus.data.RegionsModelFiQuS import RegionsModel


class Mesh:
    def __init__(self, fdm, verbose=True):
        """
        Class to preparing brep files by adding terminals.
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name

        self.geom_folder = Path(self.model_folder).parent

        self.verbose = verbose
        regions_file = os.path.join(self.geom_folder, f'{self.magnet_name}.regions')
        self.cctrm = uff.read_data_from_yaml(regions_file, RegionsModel)
        winding_info_file = os.path.join(self.geom_folder, f'{self.magnet_name}.wi')
        self.cctwi = uff.read_data_from_yaml(winding_info_file, WindingsInformation)
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()
        self.model_file = f"{os.path.join(self.model_folder, self.magnet_name)}.msh"
        self.formers = []
        self.powered_vols = []
        self.air_boundary_tags = []

    def _find_surf(self, volume_tag):
        for surf in gmsh.model.getBoundary([(3, volume_tag)], oriented=False):
            _, _, zmin, _, _, zmax = gmsh.model.occ.getBoundingBox(*surf)
            z = (zmin + zmax) / 2
            if math.isclose(z, self.cctdm.geometry.air.z_min, rel_tol=1e-5) or math.isclose(z, self.cctdm.geometry.air.z_max, rel_tol=1e-5):
                return surf[1]

    def generate_physical_groups(self, gui=False):
        if self.verbose:
            print('Generating Physical Groups Started')
            start_time = timeit.default_timer()
        r_types = ['w' for _ in self.cctwi.w_names] + ['f' for _ in self.cctwi.f_names]  # needed for picking different pow_surf later on
        vol_max_loop = 0

        for i, (f_name, r_type, r_name, r_tag) in enumerate(zip(self.cctwi.w_names + self.cctwi.f_names, r_types, self.cctrm.powered.vol.names, self.cctrm.powered.vol.numbers)):
            vol_dict = json.load(open(f"{os.path.join(self.geom_folder, f_name)}.vi"))
            volumes_file = np.array(vol_dict['all'])
            vol_max_file = np.max(volumes_file)
            vols_to_use = volumes_file + vol_max_loop
            v_tags = (list(map(int, vols_to_use)))
            vol_max_loop = vol_max_file + vol_max_loop
            self.powered_vols.extend(v_tags)  # used later for meshing
            gmsh.model.addPhysicalGroup(dim=3, tags=v_tags, tag=r_tag)
            gmsh.model.setPhysicalName(dim=3, tag=r_tag, name=r_name)
            powered_in_surf = self._find_surf(v_tags[0])
            gmsh.model.addPhysicalGroup(dim=2, tags=[powered_in_surf], tag=self.cctrm.powered.surf_in.numbers[i])
            gmsh.model.setPhysicalName(dim=2, tag=self.cctrm.powered.surf_in.numbers[i], name=self.cctrm.powered.surf_in.names[i])
            powered_out_surf = self._find_surf(v_tags[-1])
            gmsh.model.addPhysicalGroup(dim=2, tags=[powered_out_surf], tag=self.cctrm.powered.surf_out.numbers[i])
            gmsh.model.setPhysicalName(dim=2, tag=self.cctrm.powered.surf_out.numbers[i], name=self.cctrm.powered.surf_out.names[i])

        vol_max_loop = v_tags[-1]

        for r_name, r_tag in zip(self.cctrm.induced.vol.names, self.cctrm.induced.vol.numbers):
            vol_max_loop += 1
            self.formers.append(vol_max_loop)
            gmsh.model.addPhysicalGroup(dim=3, tags=[vol_max_loop], tag=r_tag)
            gmsh.model.setPhysicalName(dim=3, tag=r_tag, name=r_name)

        vol_max_loop += 1
        gmsh.model.addPhysicalGroup(dim=3, tags=[vol_max_loop], tag=self.cctrm.air.vol.number)
        gmsh.model.setPhysicalName(dim=3, tag=self.cctrm.air.vol.number, name=self.cctrm.air.vol.name)
        abt = gmsh.model.getEntities(2)[-3:]
        self.air_boundary_tags = [surf[1] for surf in abt]
        gmsh.model.addPhysicalGroup(dim=2, tags=self.air_boundary_tags, tag=self.cctrm.air.surf.number)
        gmsh.model.setPhysicalName(dim=2, tag=self.cctrm.air.surf.number, name=self.cctrm.air.surf.name)

        # air_line_tags = []
        # for air_boundary_tag in self.air_boundary_tags:
        #     air_line_tags.extend(gmsh.model.getBoundary([(2, air_boundary_tag)], oriented=False)[1])
        # self.air_center_line_tags = [int(np.max(air_line_tags) + 1)]  # this assumes that the above found the lines of air boundary but not the one in the middle that is just with the subsequent tag
        # gmsh.model.addPhysicalGroup(dim=1, tags=self.air_center_line_tags, tag=self.cctrm.air.line.number)
        # gmsh.model.setPhysicalName(dim=1, tag=self.cctrm.air.line.number, name=self.cctrm.air.line.name)

        gmsh.model.occ.synchronize()
        if gui:
            self.gu.launch_interactive_GUI()
        if self.verbose:
            print(f'Generating Physical Groups Took {timeit.default_timer() - start_time:.2f} s')

    def generate_mesh(self, gui=False):
        if self.verbose:
            print('Generating Mesh Started')
            start_time = timeit.default_timer()
        # gmsh.option.setNumber("Mesh.AngleToleranceFacetOverlap", 0.01)
        gmsh.option.setNumber("Mesh.Algorithm", 5)
        gmsh.option.setNumber("Mesh.Algorithm3D", 10)
        gmsh.option.setNumber("Mesh.AllowSwapAngle", 20)
        line_tags_transfinite = []
        num_div_transfinite = []
        line_tags_mesh: List[Any] = []
        point_tags_mesh = []
        for vol_tag in self.powered_vols:
            vol_surf_tags = gmsh.model.getAdjacencies(3, vol_tag)[1]
            for surf_tag in vol_surf_tags:
                line_tags = gmsh.model.getAdjacencies(2, surf_tag)[1]
                # line_tags_mesh.extend(line_tags)
                line_lengths = []
                for line_tag in line_tags:
                    point_tags = gmsh.model.getAdjacencies(1, line_tag)[1]
                    if line_tag not in line_tags_mesh:
                        line_tags_mesh.append(line_tag)
                    x = []
                    y = []
                    z = []
                    for p, point_tag in enumerate(point_tags):
                        xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, point_tag)
                        x.append((xmin + xmax) / 2)
                        y.append((ymin + ymax) / 2)
                        z.append((zmin + zmax) / 2)
                        if point_tag not in point_tags_mesh:
                            point_tags_mesh.append(point_tag)
                    line_lengths.append(math.sqrt((x[0] - x[1]) ** 2 + (y[0] - y[1]) ** 2 + (z[0] - z[1]) ** 2))
                    # num_div = math.ceil(dist / self.cctdm.mesh_generators.MeshSizeWindings) + 1
                l_length_min = np.min(line_lengths)
                for line_tag, l_length in zip(line_tags, line_lengths):
                    aspect = l_length / l_length_min
                    if aspect > self.cctdm.mesh.MaxAspectWindings:
                        num_div = math.ceil(l_length / (l_length_min * self.cctdm.mesh.MaxAspectWindings)) + 1
                        if line_tag in line_tags_transfinite:
                            idx = line_tags_transfinite.index(line_tag)
                            num_div_set = num_div_transfinite[idx]
                            if num_div_set < num_div:
                                num_div_transfinite[idx] = num_div
                        else:
                            line_tags_transfinite.append(line_tag)
                            num_div_transfinite.append(num_div)
        for line_tag, num_div in zip(line_tags_transfinite, num_div_transfinite):
            gmsh.model.mesh.setTransfiniteCurve(line_tag, num_div)

        gmsh.model.setColor([(1, i) for i in line_tags_mesh], 255, 0, 0)  # , recursive=True)  # Red

        # gmsh.model.mesh.setTransfiniteCurve(self.air_center_line_tags[0], 15)
        # gmsh.model.setColor([(1, i) for i in self.air_center_line_tags], 0, 0, 0)

        sld = gmsh.model.mesh.field.add("Distance")  # straight line distance
        gmsh.model.mesh.field.setNumbers(sld, "CurvesList", line_tags_mesh)
        # gmsh.model.mesh_generators.field.setNumbers(1, "PointsList", point_tags_mesh)
        gmsh.model.mesh.field.setNumber(sld, "Sampling", 100)
        slt = gmsh.model.mesh.field.add("Threshold")  # straight line threshold
        gmsh.model.mesh.field.setNumber(slt, "InField", sld)
        gmsh.model.mesh.field.setNumber(slt, "SizeMin", self.cctdm.mesh.ThresholdSizeMin)
        gmsh.model.mesh.field.setNumber(slt, "SizeMax", self.cctdm.mesh.ThresholdSizeMax)
        gmsh.model.mesh.field.setNumber(slt, "DistMin", self.cctdm.mesh.ThresholdDistMin)
        gmsh.model.mesh.field.setNumber(slt, "DistMax", self.cctdm.mesh.ThresholdDistMax)
        # gmsh.model.mesh_generators.field.add("Min", 7)
        # gmsh.model.mesh_generators.field.setNumbers(7, "FieldsList", [slt])
        gmsh.model.mesh.field.setAsBackgroundMesh(slt)
        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
        gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
        gmsh.option.setNumber("Mesh.OptimizeNetgen", 0)
        gmsh.model.mesh.generate(3)
        if self.verbose:
            print(f'Generating Mesh Took {timeit.default_timer() - start_time:.2f} s')
        if gui:
            self.gu.launch_interactive_GUI()

    def generate_cuts(self, gui=False):
        if self.verbose:
            print('Generating Cuts Started')
            start_time = timeit.default_timer()
        for vol, surf_in, surf_out in zip(self.cctrm.powered.vol.numbers, self.cctrm.powered.surf_in.numbers, self.cctrm.powered.surf_out.numbers):
            gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[vol], subdomainTags=[surf_in, surf_out], dims=[1, 2, 3])
        for vol in self.cctrm.induced.vol.numbers:
            gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[vol], dims=[1, 2, 3])
        gmsh.model.mesh.computeHomology()
        if self.verbose:
            print(f'Generating Cuts Took {timeit.default_timer() - start_time:.2f} s')
        if gui:
            self.gu.launch_interactive_GUI()

    def save_mesh(self, gui=False):
        if self.verbose:
            print('Saving Mesh Started')
            start_time = timeit.default_timer()
        gmsh.write(self.model_file)
        if self.verbose:
            print(f'Saving Mesh Took {timeit.default_timer() - start_time:.2f} s')
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def load_mesh(self, gui=False):
        gmsh.open(self.model_file)
        if gui:
            self.gu.launch_interactive_GUI()



import os
from pathlib import Path
import gmsh
import json
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as lines

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import GeometricFunctions as Func
from fiqus.utils.Utils import RoxieParsers as Pars
from fiqus.data import DataFiQuS as dF


class PostProcess:
    def __init__(self, data: dF.FDM() = None, sett: dF.FiQuSSettings() = None, solution_folder: str = None,
                 verbose: bool = False):
        """
        Class to post process results
        :param data: FiQuS data model
        :param sett: settings data model
        :param verbose: If True more information is printed in python console.
        """
        self.data: dF.FDM() = data
        self.set: dF.FiQuSSettings() = sett
        self.solution_folder = solution_folder
        self.verbose: bool = verbose

        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize()

        self.brep_iron_curves = {1: set(), 2: set(), 3: set(), 4: set()}
        self.strands = None
        self.crns = None
        self.BB_err_mean = []
        self.BB_err_min = []
        self.BB_err_max = []
        self.postprocess_parameters = dict.fromkeys(['overall_error', 'minimum_diff', 'maximum_diff'])
        self.geom_folder = os.path.dirname(os.path.dirname(self.solution_folder))
        self.model_file = f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.map2d"

        self.II = (self.set.Model_Data_GS.general_parameters.I_ref[0] if self.data.magnet.postproc.compare_to_ROXIE
                   else self.data.magnet.solve.I_initial[0])

        if self.data.magnet.postproc.plot_all != 'False':
            self.fiqus = None
            self.roxie = None
            plt.figure(1)
            self.ax = plt.axes()
            self.ax.set_xlabel('x [m]')
            self.ax.set_ylabel('y [m]')
            self.ax.set_xlim(0, 0.09)
            self.ax.set_ylim(0, 0.04)

            if self.data.magnet.postproc.compare_to_ROXIE:
                fig2 = plt.figure(2)
                self.ax2 = fig2.add_subplot(projection='3d')
                self.ax2.set_xlabel('x [m]')
                self.ax2.set_ylabel('y [m]')
                self.ax2.set_zlabel('Absolute Error [T]')
                self.fig4 = plt.figure(4)
                self.ax4 = plt.axes()
                self.ax4.set_xlabel('x [cm]')
                self.ax4.set_ylabel('y [cm]')
                self.ax4.set_aspect('equal', 'box')
            fig3 = plt.figure(3)
            self.ax3 = fig3.add_subplot(projection='3d')
            self.ax3.set_xlabel('x [m]')
            self.ax3.set_ylabel('y [m]')
            self.ax3.set_zlabel('norm(B) [T]')

        name = 'b_Omega_p'
        gmsh.open(os.path.join(self.solution_folder, f"{name}.pos"))

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def loadStrandPositions(self):
        self.strands = json.load(open(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.strs"))

    def loadHalfTurnCornerPositions(self):
        self.crns = json.load(open(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.crns"))

    def plotHalfTurnGeometry(self):
        for i in range(len(self.crns['iL'])):
            self.ax.add_line(lines.Line2D([self.crns['iL'][i][0], self.crns['iR'][i][0]],
                                          [self.crns['iL'][i][1], self.crns['iR'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['oL'][i][0], self.crns['oR'][i][0]],
                                          [self.crns['oL'][i][1], self.crns['oR'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['oR'][i][0], self.crns['iR'][i][0]],
                                          [self.crns['oR'][i][1], self.crns['iR'][i][1]], color='green'))
            self.ax.add_line(lines.Line2D([self.crns['iL'][i][0], self.crns['oL'][i][0]],
                                          [self.crns['iL'][i][1], self.crns['oL'][i][1]], color='green'))
            cc_fiqus = Func.centroid([self.crns['iL'][i][0], self.crns['iR'][i][0],
                                      self.crns['oR'][i][0], self.crns['oL'][i][0]],
                                     [self.crns['iL'][i][1], self.crns['iR'][i][1],
                                      self.crns['oR'][i][1], self.crns['oL'][i][1]])

            if self.data.magnet.postproc.compare_to_ROXIE:
                self.ax.add_line(lines.Line2D([self.crns['iLr'][i][0], self.crns['iRr'][i][0]],
                                              [self.crns['iLr'][i][1], self.crns['iRr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['oLr'][i][0], self.crns['oRr'][i][0]],
                                              [self.crns['oLr'][i][1], self.crns['oRr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['oRr'][i][0], self.crns['iRr'][i][0]],
                                              [self.crns['oRr'][i][1], self.crns['iRr'][i][1]],
                                              color='red', linestyle='dashed'))
                self.ax.add_line(lines.Line2D([self.crns['iLr'][i][0], self.crns['oLr'][i][0]],
                                              [self.crns['iLr'][i][1], self.crns['oLr'][i][1]],
                                              color='red', linestyle='dashed'))
                cc_roxie = Func.centroid(
                    [self.crns['iLr'][i][0], self.crns['iRr'][i][0], self.crns['oRr'][i][0], self.crns['oLr'][i][0]],
                    [self.crns['iLr'][i][1], self.crns['iRr'][i][1], self.crns['oRr'][i][1], self.crns['oLr'][i][1]])

        self.fiqus = self.ax.scatter(cc_fiqus[0], cc_fiqus[1], c="green")
        if self.data.magnet.postproc.compare_to_ROXIE:
            self.roxie = self.ax.scatter(cc_roxie[0], cc_roxie[1], edgecolor='r', facecolor='none')

    def postProcess(self):
        def _printState(bb):
            perc = round(bb / blocks_nr * 100)
            print("Info    : [" + f"{' ' if perc < 10 else ''}" + f"{' ' if perc < 100 else ''}" + f"{perc}" +
                  "%] Interpolating within block" + f"{str(bb)}")

        def _fetchRoxieData():
            roxie_x.append(matrix[row, 3])
            roxie_y.append(matrix[row, 4])
            strands_area.append(matrix[row, 7])
            BB_roxie_x.append(matrix[row, 5])
            BB_roxie_y.append(matrix[row, 6])
            BB_roxie.append(np.linalg.norm(np.array([matrix[row, 5], matrix[row, 6]])))

        def _probeFromView():
            BB_x = []
            BB_y = []
            BB = []
            b_list = [b for b, field in enumerate(self.data.magnet.postproc.variables) if field == 'b']
            if not b_list:
                raise Exception("The interpolation of the field at the strands locations can not be executed: "
                                "the field 'b' is not listed in 'post_processors' -> 'variables'")
            b_field_index = 0
            while self.data.magnet.postproc.volumes[b_list[b_field_index]] != 'Omega_p':
                b_field_index += 1
            # self.data.magnet.post_proc.variables.index('b')
            for ss in range(len(strands_x)):
                probe_data = gmsh.view.probe(gmsh.view.getTags()[0] if len(b_list) == 1 else b_list[b_field_index],
                                             strands_x[ss], strands_y[ss], 0)[0]
                BB_x.append(probe_data[0])
                BB_y.append(probe_data[1])
                BB.append(np.linalg.norm(np.array([BB_x[-1], BB_y[-1]])))
            return BB_x, BB_y, BB

        def _updateMap2dFile(blk_nr, ht_nr):
            with open(self.model_file, 'a') as f:
                content = []
                s = row - strands
                for x, y, Bx, By in zip(strands_x, strands_y, BB_strands_x, BB_strands_y):
                    s += 1
                    content.append(f"     {blk_nr}     {ht_nr}      {s}      {x * 1e3:.4f}       "
                                   f"{y * 1e3:.4f}    {Bx:.4f}    {By:.4f}     {1.0}   "
                                   f"{self.II / strands:.2f}  {1.0}\n")
                f.writelines(content)

        def _computeError():
            BB_err = BB_strands - np.array(BB_roxie)
            self.BB_err_mean.append(np.mean(abs(BB_err)))
            self.BB_err_min.append(np.min(abs(BB_err)))
            self.BB_err_min.append(np.max(abs(BB_err)))
            return BB_err

        def _plotData():
            map2d = None
            pos_roxie = None
            if self.data.magnet.postproc.compare_to_ROXIE:
                map2d = self.ax.scatter(roxie_x, roxie_y, edgecolor='black', facecolor='black', s=10)
                corners = conductorPositionsList[c + cond_nr].xyCorner
                for corner in range(len(corners)):
                    self.ax.scatter(corners[corner][0] / 1e3, corners[corner][1] / 1e3, edgecolor='black',
                                    facecolor='black', s=10)
                if flag_contraction:
                    self.ax.scatter([x * (1 - 0.002) for x in parser_x],
                                    [y * (1 - 0.002) for y in parser_y], c='r', s=10)
                else:
                    self.ax.scatter(parser_x, parser_y, c='r', s=10)
                self.ax2.scatter3D(roxie_x, roxie_y, BB_abs_err, c=BB_abs_err, cmap='viridis')  # , vmin=-0.2, vmax=0.2)
                self.ax4.scatter(np.array(roxie_x) * 1e2, np.array(roxie_y) * 1e2, s=1, c=np.array(BB_abs_err) * 1e3, cmap='viridis')
                pos_roxie = self.ax3.scatter3D(roxie_x, roxie_y, BB_roxie, c=BB_roxie, cmap='Reds', vmin=0, vmax=10)
            pos = self.ax3.scatter3D(strands_x, strands_y, BB_strands, c=BB_strands, cmap='Greens', vmin=0, vmax=10)
            return map2d, pos_roxie, pos

        print(f"Info    : {self.data.general.magnet_name} - I n t e r p o l a t i n g . . .")
        print("Info    : Interpolating magnetic flux density ...")

        if self.data.magnet.postproc.compare_to_ROXIE:
            roxie_path = self.data.magnet.postproc.compare_to_ROXIE  # f"C:\\Users\\avitrano\\PycharmProjects\\steam_sdk\\tests\\builders\\model_library\\magnets\\{self.data.general.magnet_name}\\input"
            flag_contraction = False
            # flag_self_field = False
            path_map2d = Path(roxie_path)
            # path_map2d = Path(roxie_path, "MQXA_All_" +
            #                   f"{'WithIron_' if self.data.magnet.geometry.with_iron_yoke else 'NoIron_'}" +
            #                   f"{'WithSelfField' if flag_self_field else 'NoSelfField'}" +
            #                   f"{'' if flag_contraction else '_no_contraction'}" + ".map2d")
            matrix = Pars.parseMap2d(map2dFile=path_map2d)

            if self.data.magnet.postproc.plot_all != 'False':
                path_cond2d = Path(os.path.join(os.path.dirname(roxie_path), self.data.general.magnet_name + ".cond2d"))
                # path_cond2d = Path(os.path.dirname(roxie_path), "MQXA_All_NoIron_NoSelfField" +
                #                    f"{'' if flag_contraction else '_no_contraction'}" + ".cond2d")
                conductorPositionsList = Pars.parseCond2d(path_cond2d)

        with open(self.model_file, 'w') as file:
            file.write("  BL.   COND.    NO.    X-POS/MM     Y-POS/MM    BX/T       BY/T"
                       "      AREA/MM**2 CURRENT FILL FAC.\n\n")

        cond_nr = 0
        row = 0
        c_nr_list = []  # list of conductors per block
        blocks_nr = self.strands['block'][-1]
        blocks, hts = np.array(self.strands['block']), np.array(self.strands['ht'])
        for blk in range(blocks_nr):
            _printState(blk + 1)
            c_nr_list.append(hts[np.where(blocks == blk + 1)[0][-1]] - sum(c_nr_list))
            for c in range(c_nr_list[blk]):
                parser_x, parser_y, roxie_x, roxie_y, strands_area, BB_roxie, BB_roxie_x, BB_roxie_y = \
                    [], [], [], [], [], [], [], []
                strands = 0
                ht = hts[row]
                while ht == c + 1 + cond_nr and row < len(self.strands['x']):
                    parser_x.append(self.strands['x'][row])
                    parser_y.append(self.strands['y'][row])
                    if self.data.magnet.postproc.compare_to_ROXIE:
                        _fetchRoxieData()
                    strands += 1
                    row += 1
                    if row < len(self.strands['x']) - 1:
                        ht = hts[row]

                strands_x = roxie_x if self.data.magnet.postproc.compare_to_ROXIE else parser_x
                strands_y = roxie_y if self.data.magnet.postproc.compare_to_ROXIE else parser_y

                BB_strands_x, BB_strands_y, BB_strands = _probeFromView()

                _updateMap2dFile(blk + 1, c + 1 + cond_nr)

                if self.data.magnet.postproc.compare_to_ROXIE:
                    BB_abs_err = _computeError()

                if self.data.magnet.postproc.plot_all != 'False':  # and (blk == 0 or blk == 1):
                    if self.data.magnet.postproc.compare_to_ROXIE:
                        map2d_strands, scatter3D_pos_roxie, scatter3D_pos = _plotData()
                    else:
                        scatter3D_pos = _plotData()[2]

            cond_nr += c_nr_list[blk]

        if self.data.magnet.postproc.compare_to_ROXIE:
            self.postprocess_parameters['overall_error'] = np.linalg.norm(self.BB_err_mean)
            self.postprocess_parameters['minimum_diff'] = np.min(self.BB_err_mean)
            self.postprocess_parameters['maximum_diff'] = np.max(self.BB_err_mean)

        print(f"Info    : {self.data.general.magnet_name} - E n d   I n t e r p o l a t i n g")

        if self.data.magnet.postproc.plot_all != 'False':
            self.plotHalfTurnGeometry()
            if self.data.magnet.postproc.compare_to_ROXIE:
                cax4 = self.fig4.add_axes([self.ax4.get_position().x1 + 0.02, self.ax4.get_position().y0,
                                           0.02, self.ax4.get_position().height])
                cbar = plt.colorbar(self.ax4.get_children()[2], cax=cax4)
                cbar.ax.set_ylabel('Absolute error [mT]', rotation=270)
                self.ax3.legend([scatter3D_pos, scatter3D_pos_roxie], ['FiQuS', 'ROXIE'], numpoints=1)
                # pickle.dump(fig, open('Bfield.fig.pickle', 'wb'))
                self.ax.legend([self.fiqus, self.roxie, map2d_strands], ['FiQuS', 'ROXIEparser', 'ROXIE'], numpoints=1)

            self.fig4.savefig(f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.svg", bbox_inches='tight')

            if self.data.magnet.postproc.plot_all == 'True':
                plt.show()

        # os.remove(os.path.join(self.solution_folder, 'b_Omega_p.pos'))
        # os.remove(f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.pre")
        # os.remove(f"{os.path.join(os.path.dirname(self.solution_folder), self.data.general.magnet_name)}.msh")



import os
import copy
import math
import collections
import logging

import json
import timeit
import numpy as np
import gmsh

from fiqus.data import RegionsModelFiQuS as Reg_Mod_FiQ
from fiqus.data.DataWindingsCCT import WindingsInformation  # for winding information
from fiqus.utils.Utils import FilesAndFolders as uff
from fiqus.utils.Utils import GmshUtils

logger = logging.getLogger(__name__)
class Hexahedrons:

    def __init__(self):
        """
        Points generator for CCT windings to give hexahedron volumes for hexahedron meshing in Gmsh.
        """
        self.vertices_to_surf = [[0, 4, 7, 3], [1, 5, 6, 2], [1, 5, 4, 0], [2, 6, 7, 3], [0, 1, 2, 3], [4, 5, 6, 7]]
        # Node ordering following Gmsh for hexahedron https://gmsh.info/doc/texinfo/gmsh.html#Legacy-formats --> node ordering
        # above _vertices_to_surf is only used for manipulation of the hexahedra points on the data model level. Gmsh uses points connections and line connections for creating surfaces

        self.hexes = {}
        self._corner_points = {}
        self._sur_names = ['x-', 'x+', 'y-', 'y+', 'z-', 'z+']
        self._op_sign = {'+': '-', '-': '+'}
        self._sign = {1: '+', -1: '-'}

    def _add_corner_points(self, corner_num, x, y, z, ct):
        self._corner_points[corner_num] = {'x': x, 'y': y, 'z': z, 'ct': ct}  # ct is channel turn

    def _generate_hexes(self):
        for h in range(len(self._corner_points[0]['x'])):
            self.hexes[h] = {}
            for p_num, p_coor in self._corner_points.items():
                self.hexes[h][p_num] = {}
                for coor in ['x', 'y', 'z']:
                    self.hexes[h][p_num][coor] = p_coor[coor][h]
            self.hexes[h]['ct'] = p_coor['ct'][h]
        return p_coor['ct'][h]

    def _add_elem_to_elem(self, start_elem, dir_str, dist):
        """
        :param start_elem: idx of start element
        :param dir_str: direction of adding element, options are: 'x-', 'x+', 'y-', 'y+', 'z-', 'z+'
        :param dist: length of new element in the direction specified above [m]
        :return: index of created element
        """
        op_dir = dir_str[0] + self._op_sign[dir_str[1]]
        new_elem = start_elem + math.copysign(1, start_elem)
        self.hexes[new_elem] = {}
        source_points = self.vertices_to_surf[self._sur_names.index(dir_str)]
        destination_points = self.vertices_to_surf[self._sur_names.index(op_dir)]
        for d_p, s_p in zip(destination_points + source_points, source_points + source_points):
            self.hexes[new_elem][d_p] = copy.deepcopy(self.hexes[start_elem][s_p])
        for s_p in source_points:  # modif_points:
            self.hexes[new_elem][s_p][dir_str[0]] = self.hexes[new_elem][s_p][dir_str[0]] + dist
        self.hexes[new_elem]['ct'] = self.hexes[start_elem]['ct']
        return new_elem

    def _reorder_and_renumber_hexes(self):
        self.hexes = collections.OrderedDict(sorted(self.hexes.items()))
        first_hex_num = list(self.hexes.keys())[0]
        temp_dict = {}
        for h_num, h_dict in self.hexes.items():
            temp_dict[h_num - first_hex_num + 1] = h_dict
        self.hexes = temp_dict
        for h_num, h_dict in self.hexes.items():
            turn = h_dict.pop('ct', None)
            self.hexes[h_num] = collections.OrderedDict(sorted(h_dict.items()))
            self.hexes[h_num]['ct'] = turn


class Winding(Hexahedrons):

    def __init__(self, cctdm, layer, post_proc=False):

        super(Winding, self).__init__()
        self.cctdm = cctdm
        self.layer = layer
        self.post_proc = post_proc
        self.z_corr = 0
        self._populate_corners()
        self.z_corr = self._center_z_and_offset()
        self._generate_hexes()

    def _populate_corners(self):
        r"""
        Generates hexahedron with 8 corners numbered 0-7. The size of hexahedron is taken from groove size of cct as specified in yaml. The length is from number of elements per turn.
               y (v)
            3----------2
            |\     ^   |\
            | \    |   | \
            |  \   |   |  \
            |   7------+---6
            |   |  +-- |-- | -> x (u)
            0---+---\--1   |
             \  |    \  \  |
              \ |     \  \ |
               \|      z(w)\|
                4----------5

        :return: none
        """
        self._sign_of_rot = math.copysign(1, self.cctdm.geometry.windings.alphas[self.layer])
        if self._sign_of_rot > 0:
            corner_seq = {
                0: {'w_s': -1, 'h_s': -1, 'far': False},  # 'blcf'
                1: {'w_s': -1, 'h_s': 1, 'far': False},  # 'brcf'
                2: {'w_s': 1, 'h_s': 1, 'far': False},  # 'trcf'
                3: {'w_s': 1, 'h_s': -1, 'far': False},  # 'tlcf'
                4: {'w_s': -1, 'h_s': -1, 'far': True},  # bottom left corner close   #'blcc'
                5: {'w_s': -1, 'h_s': 1, 'far': True},  # bottom right corner close      #'brcc'
                6: {'w_s': 1, 'h_s': 1, 'far': True},  # top right corner close     #'trcc'
                7: {'w_s': 1, 'h_s': -1, 'far': True},  # top right corner close  #'tlcc'
            }
        else:
            corner_seq = {
                0: {'w_s': 1, 'h_s': 1, 'far': False},  # 'trcf'
                1: {'w_s': 1, 'h_s': -1, 'far': False},  # 'brcf'
                2: {'w_s': -1, 'h_s': -1, 'far': False},  # 'blcf'
                3: {'w_s': -1, 'h_s': 1, 'far': False},  # 'tlcf'
                4: {'w_s': 1, 'h_s': 1, 'far': True},  # top right corner close     #'trcc'
                5: {'w_s': 1, 'h_s': -1, 'far': True},  # bottom right corner close  #'brcc'
                6: {'w_s': -1, 'h_s': -1, 'far': True},  # bottom left corner close   #'blcc'
                7: {'w_s': -1, 'h_s': 1, 'far': True},  # top left corner close      #'tlcc'
            }
        for corner_num, corner_dict in corner_seq.items():
            self._add_corner_points(corner_num, *self.calc_turns_corner_coords(corner_dict['w_s'], corner_dict['h_s'], far=corner_dict['far']))

    def calc_turns_corner_coords(self, w_s=1, h_s=1, far=False):
        """
        Based on https://doi.org/10.1016/j.cryogenics.2020.103041
        :param h_s: wire height sign
        :param w_s: wire width sign
        :param far: for far surface set to True, for close surface set to False. This refers to walls of hexagon.
        :return:
        """
        r_wm = self.cctdm.geometry.windings.r_wms[self.layer]  # radius of winding layer, in the middle of the conductor groove [m]
        w_i = self.cctdm.geometry.windings.lps[self.layer]  # layer pitch [m]
        n_turns = self.cctdm.geometry.windings.n_turnss[self.layer]  # number of turns [-]
        nept = self.cctdm.geometry.windings.ndpts[self.layer] - 1  # number of elements per turn [-]
        alpha = self.cctdm.geometry.windings.alphas[self.layer] * np.pi / 180  # inclination of the winding [deg]
        theta_0 = 0  # offset start at theta_0, i.e. angular position of the beginning of the 1st turn[deg]
        z_0 = 0  # offset start at z_0, i.e. axial position of the beginning of the 1st turn [m]
        hh = 0.5 * h_s * self.cctdm.geometry.windings.wwhs[self.layer]  # half of h (groove height)
        hw = 0.5 * w_s * self.cctdm.geometry.windings.wwws[self.layer]  # half of w (groove width)
        tot_angle = 2 * np.pi * (n_turns + 1 / nept)
        tot_points = int(nept * n_turns) + 1
        if alpha > 0:
            theta = np.linspace(0, tot_angle, tot_points, endpoint=False)
        else:
            theta = np.linspace(0 + np.pi, tot_angle + np.pi, tot_points, endpoint=False)  # this rotates terminals for opposite inclination to be on the opposite side
        if far:
            theta = theta[1:]  # give all points but the first one, so the far surface can be created
        else:
            theta = theta[:-1]  # give all the points but the last one so a near surface can be created
        if theta.size < 2:
            raise ValueError(f'Combination of number of division points per turn (ndpts) and number of turns (n_turnss) must result in at least 2 Hexahedrons, but inputs result in only {theta.size}!')
        C = r_wm * np.arctan(alpha) * np.cos(theta - theta_0) + w_i / (2 * np.pi)
        D = np.sqrt(r_wm ** 2 + np.square(C))
        # --  x, y, z points coordinates
        xs = (r_wm + hh) * np.cos(theta) - np.sin(theta) * hw * C / D
        ys = (r_wm + hh) * np.sin(theta) + np.cos(theta) * hw * C / D
        zs = r_wm * np.sin(theta - theta_0) / np.tan(alpha) + w_i * (theta - theta_0) / (2 * np.pi) + z_0 - hw * r_wm / D  # original formula
        if alpha > 0:
            channel_turn = np.ceil((theta - 2 * np.pi / (nept + 1)) / (2 * np.pi))
            # channel_turn[0] = channel_turn[1]  # make the hex at theta 0 to belong to the first turn.
        else:
            channel_turn = np.ceil((theta - np.pi - 2 * np.pi / (nept + 1)) / (2 * np.pi))
        channel_turn[0] = channel_turn[1]  # make the hex at theta 0 to belong to the first turn.
        return xs.tolist(), ys.tolist(), zs.tolist(), list(map(int, channel_turn))

    def _center_z_and_offset(self, offset=0):
        z_mins = []
        z_maxs = []
        for c_val in self._corner_points.values():
            z_mins.append(np.min(c_val['z']))
            z_maxs.append(np.max(c_val['z']))
        z_min_wind = np.min(np.array(z_mins))
        z_max_wind = np.max(np.array(z_maxs))
        z_centre = (z_min_wind + z_max_wind) / 2
        if not self.post_proc:
            for c_key, c_val in self._corner_points.items():
                self._corner_points[c_key]['z'] = (c_val['z'] - z_centre + offset).tolist()
        z_mins = []
        z_maxs = []
        for c_val in self._corner_points.values():
            z_mins.append(np.min(c_val['z']))
            z_maxs.append(np.max(c_val['z']))
        self.z_min_winding_layer = np.min(np.array(z_mins))
        self.z_max_winding_layer = np.max(np.array(z_maxs))
        return z_centre + offset


class FQPL(Hexahedrons):
    def __init__(self, cctdm, layer):
        """
        FQPL hex generator
        :param cctdm: cct data model object parsed from yaml input file
        :param layer:   loop number, integer
        """
        super(FQPL, self).__init__()
        self.cctdm = cctdm
        self.layer = layer
        self._calc_z_unit_len()
        self._populate_corners()
        self._generate_hexes()
        self._generate_fqpl()
        self._rotate_fqpl()
        self._reorder_and_renumber_hexes()
        self._clean_attr()

    def _calc_z_unit_len(self):
        if self.cctdm.geometry.fqpls.z_starts[self.layer] == 'z_min':
            self.near = self.cctdm.geometry.air.z_min
            self.z_unit_len = (self.cctdm.geometry.fqpls.z_ends[self.layer] - self.near - self.cctdm.geometry.fqpls.fwws[self.layer]) / self.cctdm.geometry.fqpls.fndpls[self.layer]

        elif self.cctdm.geometry.fqpls.z_starts[self.layer] == 'z_max':
            self.near = self.cctdm.geometry.air.z_max
            self.z_unit_len = (self.near - self.cctdm.geometry.fqpls.z_ends[self.layer] - self.cctdm.geometry.fqpls.fwws[self.layer]) / self.cctdm.geometry.fqpls.fndpls[self.layer]
        else:
            raise ValueError(f'fqpl.z_starts parameter must be a string equal to z_min or z_max, but {self.cctdm.geometry.fqpls.z_starts[self.layer]} was given!')

        self.extrusion_sign = math.copysign(1, self.cctdm.geometry.fqpls.z_ends[self.layer] - self.near)

    def _calc_first_corner_coords(self, w_s, h_s, far, offset):
        x = [self.cctdm.geometry.fqpls.r_ins[self.layer] + self.cctdm.geometry.fqpls.fwhs[self.layer] * h_s]  # not rotated coordinate
        y = [self.cctdm.geometry.fqpls.fwws[self.layer] * w_s]  # not rotated coordinate
        if far:
            z = [self.near - self.z_unit_len + offset]
        else:
            z = [self.near + offset]
        return x, y, z, [1]

    def _populate_corners(self):
        r"""
        Generates hexahedron with 8 corners numbered 0-7. The size of hexahedron is taken from groove size of cct as specified in yaml. The length is from number of elements per turn.
               y (v)
            3----------2
            |\     ^   |\
            | \    |   | \
            |  \   |   |  \
            |   7------+---6
            |   |  +-- |-- | -> x (u)
            0---+---\--1   |
             \  |    \  \  |
              \ |     \  \ |
               \|      z(w)\|
                4----------5

        :return: none
        """
        if self.extrusion_sign > 0:
            offset = self.z_unit_len
            corner_seq = {
                0: {'w_s': -0.5, 'h_s': 0, 'far': True},  # 'blcf'
                1: {'w_s': -0.5, 'h_s': 1, 'far': True},  # 'tlcf'
                2: {'w_s': 0.5, 'h_s': 1, 'far': True},  # 'trcf'
                3: {'w_s': 0.5, 'h_s': 0, 'far': True},  # 'brcf'
                4: {'w_s': -0.5, 'h_s': 0, 'far': False},  # bottom left corner close   #'blcc'
                5: {'w_s': -0.5, 'h_s': 1, 'far': False},  # top left corner close      #'tlcc'
                6: {'w_s': 0.5, 'h_s': 1, 'far': False},  # top right corner close     #'trcc'
                7: {'w_s': 0.5, 'h_s': 0, 'far': False},  # bottom right corner close  #'brcc'
            }
        else:
            offset = 0
            corner_seq = {
                0: {'w_s': -0.5, 'h_s': 0, 'far': False},  # 'blcf'
                1: {'w_s': -0.5, 'h_s': 1, 'far': False},  # 'tlcf'
                2: {'w_s': 0.5, 'h_s': 1, 'far': False},  # 'trcf'
                3: {'w_s': 0.5, 'h_s': 0, 'far': False},  # 'brcf'
                4: {'w_s': -0.5, 'h_s': 0, 'far': True},  # bottom left corner close   #'blcc'
                5: {'w_s': -0.5, 'h_s': 1, 'far': True},  # top left corner close      #'tlcc'
                6: {'w_s': 0.5, 'h_s': 1, 'far': True},  # top right corner close     #'trcc'
                7: {'w_s': 0.5, 'h_s': 0, 'far': True},  # bottom right corner close  #'brcc'
            }
        for corner_num, corner_dict in corner_seq.items():
            self._add_corner_points(corner_num, *self._calc_first_corner_coords(corner_dict['w_s'], corner_dict['h_s'], corner_dict['far'], offset))

    def _get_ref_point(self, elem_number, surf_n_dir_str, coord, dist):
        """
        Gets center point for a surface 'pointing' in the direction surf_n_dir_str of hex with elem_number. It then moves that point into along coordinate coord and by distance dist
        :param elem_number: hex number to take the surface from
        :param surf_n_dir_str: 'direction of surface' (kind of normal), options are: 'x-', 'x+', 'y-', 'y+', 'z-', 'z+'
        :param coord: direction to more the point in, options are: 'x', 'y', 'z'
        :param dist: distance to move the point, this is positive or negative float
        :return:
        """
        source_points = self.vertices_to_surf[self._sur_names.index(surf_n_dir_str)]  # get the indexes 4 points for the surface points in direction surf_n_dir_str
        points_coor = [self.hexes[elem_number][point] for point in source_points]  # get the points, they contain coordinates
        xs = []
        ys = []
        zs = []
        for point in points_coor:  # get coordinates into list for each axis
            for cor_list, coor in zip([xs, ys, zs], ['x', 'y', 'z']):
                cor_list.append(point[coor])
        p = {'x': np.mean(xs), 'y': np.mean(ys), 'z': np.mean(zs)}  # get averages of each axis coordinate, giving the center of the surface
        p[coord] = p[coord] + dist
        return p

    def _add_elem_with_rot(self, elem_number, surf_n_dir_str, add_ax, angle, ref_p_coor):
        """
        :param elem_number: idx of start element
        :param surf_n_dir_str: direction of adding element, options are: 'x-', 'x+', 'y-', 'y+', 'z-', 'z+'
        :param ref_p_coor: length of new element in the direction specified above [m]
        :return: index of created element
        """
        mod_coor = [surf_n_dir_str[0], add_ax]
        # mod_coor.remove(surf_n_dir_str[0])  # get the other two coordinates that arenet the direction of the surface to take

        angle_rad = np.deg2rad(angle)
        op_dir = surf_n_dir_str[0] + self._op_sign[surf_n_dir_str[1]]
        new_elem = elem_number + math.copysign(1, elem_number)
        self.hexes[new_elem] = {}
        source_points = self.vertices_to_surf[self._sur_names.index(surf_n_dir_str)]
        destination_points = self.vertices_to_surf[self._sur_names.index(op_dir)]
        for d_p, s_p in zip(destination_points + source_points, source_points + source_points):
            self.hexes[new_elem][d_p] = copy.deepcopy(self.hexes[elem_number][s_p])
        for s_p in source_points:  # modif_points:
            v1 = self.hexes[new_elem][s_p][mod_coor[0]]
            v2 = self.hexes[new_elem][s_p][mod_coor[1]]
            v1_ofset = ref_p_coor[mod_coor[0]]
            v2_ofset = ref_p_coor[mod_coor[1]]
            v1 = v1 - v1_ofset
            v2 = v2 - v2_ofset
            v1_new = np.cos(angle_rad) * v1 - np.sin(angle_rad) * v2
            v2_new = np.sin(angle_rad) * v1 + np.cos(angle_rad) * v2
            v1 = v1_new + v1_ofset
            v2 = v2_new + v2_ofset
            self.hexes[new_elem][s_p][mod_coor[0]] = v1
            self.hexes[new_elem][s_p][mod_coor[1]] = v2
        self.hexes[new_elem]['ct'] = self.hexes[elem_number]['ct']
        return new_elem

    def _generate_fqpl(self):
        new_elem = self._add_elem_to_elem(0, 'z+', self.extrusion_sign * self.z_unit_len)
        for _ in range(self.cctdm.geometry.fqpls.fndpls[self.layer] - 2):
            new_elem = self._add_elem_to_elem(new_elem, 'z+', self.extrusion_sign * self.z_unit_len)
        coord = 'x'
        ref_point = self._get_ref_point(new_elem, 'z+', coord, self.cctdm.geometry.fqpls.r_bs[self.layer])  # centre of rotation pint.
        n_seg = 10
        angle = 180 / n_seg
        for _ in range(n_seg):
            new_elem = self._add_elem_with_rot(new_elem, 'z+', coord, self.extrusion_sign * angle, ref_point)
        for _ in range(self.cctdm.geometry.fqpls.fndpls[self.layer]):
            new_elem = self._add_elem_to_elem(new_elem, 'z+', -self.extrusion_sign * self.z_unit_len)

    def _rotate_fqpl(self):
        for h_num, h_dict in self.hexes.items():
            channel_turn = h_dict.pop('ct', None)
            for p_num, p_dict in h_dict.items():
                xx = p_dict['x']
                yy = p_dict['y']
                x = xx * np.cos(np.deg2rad(self.cctdm.geometry.fqpls.thetas[self.layer])) - yy * np.sin(np.deg2rad(self.cctdm.geometry.fqpls.thetas[self.layer]))  # rotated coordinate
                y = xx * np.sin(np.deg2rad(self.cctdm.geometry.fqpls.thetas[self.layer])) + yy * np.cos(np.deg2rad(self.cctdm.geometry.fqpls.thetas[self.layer]))  # rotated coordinate
                self.hexes[h_num][p_num]['x'] = x
                self.hexes[h_num][p_num]['y'] = y
            self.hexes[h_num]['ct'] = channel_turn

    def _clean_attr(self):
        del self.cctdm
        del self.layer
        del self.near
        del self.z_unit_len


class Generate_BREPs:
    def __init__(self, fdm, verbose=True):
        """
        Class to generate (build) Windings and Formers (WF) of a canted cosine theta (CCT) magnet and save them to brep files
        :param fdm: fiqus data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name

        self.verbose = verbose
        self.cctwi = WindingsInformation()
        self.cctwi.magnet_name = self.magnet_name
        self.transl_dict = {'vol': '', 'surf': '_bd', 'surf_in': '_in', 'surf_out': '_out', 'cochain': '_cut'}
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()
        self.formers = []
        self.air = []
        gmsh.option.setString('Geometry.OCCTargetUnit', 'M')

    def generate_windings_or_fqpls(self, type_str):
        """
        Generates windings brep files, as many as number of entries in the yaml file for windings.
        :param type_str: What to generate, options are: 'windings' or 'fqpls'
        :return: None, saves brep to disk
        """
        if self.verbose:
            print(f'Generating {type_str} started')
            start_time = timeit.default_timer()

        def generate_geom(worf_in, names_in):
            worf_tags_dict_in = {}
            for worf_num, worf_in in enumerate(worf_in):
                worf_tags_dict_in[worf_num] = self.generate_hex_geom(worf_in)
                gmsh.model.occ.synchronize()
                gmsh.write(os.path.join(self.model_folder, f'{names_in[worf_num]}.brep'))
                gmsh.clear()
            return worf_tags_dict_in

        if self.cctdm.geometry.windings.names and type_str == 'windings':
            worf = [Winding(self.cctdm, i) for i, _ in enumerate(self.cctdm.geometry.windings.names)]        # worf = winding or fqpl
            names = [name for name in self.cctdm.geometry.windings.names]
            for w in worf:
                if self.cctdm.geometry.air.z_min > w.z_min_winding_layer or w.z_max_winding_layer > self.cctdm.geometry.air.z_max:
                    raise ValueError(f'{self.cctdm.geometry.air.name} region dimensions are too s mall to fit the windings turns. Please extend z_min and/or z_max')
                else:
                    pass
            worf_tags_dict = generate_geom(worf, names)

        if self.cctdm.geometry.fqpls.names and type_str == 'fqpls':
            worf = [FQPL(self.cctdm, i) for i, _ in enumerate(self.cctdm.geometry.fqpls.names)]  # worf = winding or fqpl
            names = [name for name in self.cctdm.geometry.fqpls.names]
            worf_tags_dict = generate_geom(worf, names)

        w_z_maxs = []
        w_z_mins = []
        if type_str == 'windings':
            for winding in worf:
                w_z_maxs.append(winding.z_max_winding_layer)
                w_z_mins.append(winding.z_min_winding_layer)
            self.cctwi.windings_avg_length = float(np.mean(w_z_maxs) - np.mean(w_z_mins))
            self.cctwi.windings.names = names
            self.cctwi.windings.t_in.vol_st = []
            self.cctwi.windings.t_in.surf_st = []
            self.cctwi.windings.t_out.vol_st = []
            self.cctwi.windings.t_out.surf_st = []
            self.cctwi.windings.t_in.vol_et = []
            self.cctwi.windings.t_in.surf_et = []
            self.cctwi.windings.t_out.vol_et = []
            self.cctwi.windings.t_out.surf_et = []
            self.cctwi.windings.t_in.lc_et = []
            self.cctwi.windings.t_out.lc_et = []
            self.cctwi.windings.t_in.ndpterms = self.cctdm.geometry.windings.ndpt_ins
            self.cctwi.windings.t_out.ndpterms = self.cctdm.geometry.windings.ndpt_outs
            self.cctwi.windings.t_in.z_air = self.cctdm.geometry.air.z_min
            self.cctwi.windings.t_out.z_air = self.cctdm.geometry.air.z_max
            for winding_num, winding_dict in worf_tags_dict.items():
                first_vol = list(winding_dict.keys())[0]
                last_vol = list(winding_dict.keys())[-1]
                self.cctwi.windings.t_in.vol_st.append(winding_dict[first_vol]['v_t'])
                self.cctwi.windings.t_in.surf_st.append(winding_dict[first_vol]['sf_ts'][0])
                self.cctwi.windings.t_out.vol_st.append(winding_dict[last_vol]['v_t'])
                self.cctwi.windings.t_out.surf_st.append(winding_dict[last_vol]['sf_ts'][2])
                self.cctwi.windings.t_in.vol_et.append(winding_dict[first_vol]['v_t'])
                self.cctwi.windings.t_in.surf_et.append(winding_dict[first_vol]['sf_ts'][0])
                self.cctwi.windings.t_out.vol_et.append(winding_dict[last_vol]['v_t'])
                self.cctwi.windings.t_out.surf_et.append(winding_dict[last_vol]['sf_ts'][2])

            lc_st_neg_alpha = [[1, 0], [0, 3], [3, 2], [2, 1], [6, 4], [4, 5], [5, 7], [7, 6], [1, 6], [0, 4], [3, 5], [2, 7]]
            lc_st_pos_alpha = [[3, 1], [1, 0], [0, 2], [2, 3], [7, 4], [4, 5], [5, 6], [6, 7], [3, 7], [1, 4], [0, 5], [2, 6]]
            lc_et_neg_alpha = [[5, 7], [7, 6], [6, 4], [4, 5], [1, 3], [3, 2], [2, 0], [0, 1], [5, 1], [7, 3], [6, 2], [4, 0]]
            lc_et_pos_alpha = [[5, 7], [7, 6], [6, 4], [4, 5], [1, 3], [3, 2], [2, 0], [0, 1], [5, 1], [7, 3], [6, 2], [4, 0]]
            # [[1, 2], [2, 3], [3, 4], [4, 1], [5, 6], [6, 7], [7, 8], [8, 5], [1, 5], [2, 6], [3, 7], [4, 8]]
            # [[2, 1], [1, 4], [4, 3], [3, 2], [6, 5], [5, 8], [8, 7], [7, 6], [2, 6], [1, 5], [4, 8], [3, 7]]

            lc_et_pos_corr = [[1, 0], [0, 3], [3, 2], [2, 1], [5, 4], [4, 7], [7, 6], [6, 5], [1, 5], [0, 4], [3, 7], [2, 6]]

            self.cctwi.windings.t_in.lc_st = []
            self.cctwi.windings.t_out.lc_st = []
            for alpha in self.cctdm.geometry.windings.alphas:
                if alpha > 0:
                    self.cctwi.windings.t_in.lc_st.append(lc_st_pos_alpha)
                    self.cctwi.windings.t_out.lc_st.append(lc_st_pos_alpha)
                    self.cctwi.windings.t_in.lc_et.append(lc_et_pos_corr)
                    self.cctwi.windings.t_out.lc_et.append(lc_et_pos_alpha)
                else:
                    self.cctwi.windings.t_in.lc_st.append(lc_st_neg_alpha)
                    self.cctwi.windings.t_out.lc_st.append(lc_st_neg_alpha)
                    self.cctwi.windings.t_in.lc_et.append(lc_et_neg_alpha)
                    self.cctwi.windings.t_out.lc_et.append(lc_et_pos_corr)

        if self.verbose:
            print(f'Generating {type_str} took {timeit.default_timer() - start_time:.2f} s')

    def save_volume_info(self):
        if self.cctdm.geometry.fqpls.names:
            fqpls = self.cctdm.geometry.fqpls.names
        else:
            fqpls = []

        self.cctwi.w_names = self.cctdm.geometry.windings.names
        self.cctwi.f_names = fqpls
        self.cctwi.formers = self.cctdm.geometry.formers.names
        self.cctwi.air = self.cctdm.geometry.air.name

        volume_info_file = os.path.join(self.model_folder, f'{self.cctwi.magnet_name}.wi')
        uff.write_data_to_yaml(volume_info_file, self.cctwi.dict())

    def generate_formers(self):
        if self.cctdm.geometry.formers.r_ins:
            if self.verbose:
                print('Generating Formers Started')
                start_time = timeit.default_timer()

            for f_i, _ in enumerate(self.cctdm.geometry.formers.r_ins):
                z = self.cctdm.geometry.formers.z_mins[f_i]
                dz = self.cctdm.geometry.formers.z_maxs[f_i] - self.cctdm.geometry.formers.z_mins[f_i]
                cylin_out = (3, gmsh.model.occ.addCylinder(0, 0, z, 0, 0, dz, self.cctdm.geometry.formers.r_outs[f_i], angle=2 * math.pi))  # add cylinder to align to existing bodies in the pro file
                cylin_in = (3, gmsh.model.occ.addCylinder(0, 0, z, 0, 0, dz, self.cctdm.geometry.formers.r_ins[f_i], angle=2 * math.pi))  # add another cylinder to subtract from the first one to make a tube
                cylinder = gmsh.model.occ.cut([cylin_out], [cylin_in], removeObject=True)  # subtract cylinders to make a tube
                self.formers.append(cylinder[0][0])  # keep just the tag and append
                gmsh.model.occ.synchronize()
                gmsh.write(os.path.join(self.model_folder, f'{self.cctdm.geometry.formers.names[f_i]}.brep'))
                gmsh.clear()
            if self.verbose:
                print(f'Generating formers took {timeit.default_timer() - start_time:.2f} s')

    def generate_air(self):
        if self.cctdm.geometry.air.ar:
            if self.verbose:
                print('Generating air started')
                start_time = timeit.default_timer()
            if self.cctdm.geometry.air.sh_type == 'cylinder':
                self.air = [(3, gmsh.model.occ.addCylinder(0, 0, self.cctdm.geometry.air.z_min, 0, 0, self.cctdm.geometry.air.z_max - self.cctdm.geometry.air.z_min, self.cctdm.geometry.air.ar, angle=2 * math.pi))]
            elif self.cctdm.geometry.air.sh_type == 'cuboid':
                air_box_size = [-self.cctdm.geometry.air.ar / 2, -self.cctdm.geometry.air.ar / 2, self.cctdm.geometry.air.z_min, self.cctdm.geometry.air.ar, self.cctdm.geometry.air.ar,
                                self.cctdm.geometry.air.z_max - self.cctdm.geometry.air.z_min]  # list of box size with: x, y, z, dx, dy, dz
                self.air = [(3, gmsh.model.occ.addBox(*air_box_size))]
            else:
                raise ValueError(f'Shape type: {self.cctdm.geometry.air.sh_type} is not supported!')
            gmsh.model.occ.synchronize()
            gmsh.write(os.path.join(self.model_folder, f'{self.cctdm.geometry.air.name}.brep'))
            gmsh.clear()
            if self.verbose:
                print(f'Generating air took {timeit.default_timer() - start_time:.2f} s')

    def generate_regions_file(self):
        if self.verbose:
            print('Generating Regions File Started')
            start_time = timeit.default_timer()
        cctrm = Reg_Mod_FiQ.RegionsModel()
        vrt = 1000000  # volume region tag start
        srt = 2000000  # surface region tag start
        lrt = 3000000  # line region tag start
        # -------- powered ----------
        # volumes
        cctrm.powered.vol.names = [name + self.transl_dict['vol'] for name in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names]
        cctrm.powered.vol.currents = self.cctdm.solve.windings.currents + self.cctdm.solve.fqpls.currents
        cctrm.powered.vol.sigmas = self.cctdm.solve.windings.sigmas + self.cctdm.solve.fqpls.sigmas
        cctrm.powered.vol.mu_rs = self.cctdm.solve.windings.mu_rs + self.cctdm.solve.fqpls.mu_rs
        reg = []
        for _ in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names:
            reg.append(vrt)
            vrt += 1
        cctrm.powered.vol.numbers = reg
        # surfaces
        cctrm.powered.surf_in.names = [name + self.transl_dict['surf_in'] for name in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names]
        cctrm.powered.surf_out.names = [name + self.transl_dict['surf_out'] for name in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names]
        reg = []
        for _ in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names:
            reg.append(srt)
            srt += 1
        cctrm.powered.surf_in.numbers = reg
        reg = []
        for _ in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names:
            reg.append(srt)
            srt += 1
        cctrm.powered.surf_out.numbers = reg

        # -------- induced ----------
        # volumes
        cctrm.induced.vol.names = [name + self.transl_dict['vol'] for name in self.cctdm.geometry.formers.names]
        cctrm.induced.vol.sigmas = self.cctdm.solve.formers.sigmas
        cctrm.induced.vol.mu_rs = self.cctdm.solve.formers.mu_rs
        reg = []
        for _ in self.cctdm.geometry.formers.names:
            reg.append(vrt)
            vrt += 1
        cctrm.induced.vol.numbers = reg

        # -------- air ----------
        # volumes
        cctrm.air.vol.name = self.cctdm.geometry.air.name + self.transl_dict['vol']
        cctrm.air.vol.sigma = self.cctdm.solve.air.sigma
        cctrm.air.vol.mu_r = self.cctdm.solve.air.mu_r
        cctrm.air.vol.number = vrt
        vrt += 1

        # surface
        cctrm.air.surf.name = self.cctdm.geometry.air.name + self.transl_dict['surf']
        cctrm.air.surf.number = srt
        srt += 1

        # # center line
        # cctrm.air.line.name = 'Center_line'
        # cctrm.air.line.number = lrt
        # lrt += 1
        lrt = srt

        # --------- cuts -------
        # these need to be done at the end with the highest surface tags
        cctrm.powered.cochain.names = [name + self.transl_dict['cochain'] for name in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names]
        reg = []
        for _ in self.cctdm.geometry.windings.names + self.cctdm.geometry.fqpls.names:
            reg.append(lrt)
            lrt += 1
        cctrm.powered.cochain.numbers = reg
        # induced cuts
        cctrm.induced.cochain.names = [name + self.transl_dict['cochain'] for name in self.cctdm.geometry.formers.names]
        reg = []
        for _ in self.cctdm.geometry.formers.names:
            reg.append(lrt)
            lrt += 1
        cctrm.induced.cochain.numbers = reg

        uff.write_data_to_yaml(os.path.join(self.model_folder, f'{self.cctwi.magnet_name}.regions'), cctrm.dict())

        if self.verbose:
            print(f'Generating Regions File Took {timeit.default_timer() - start_time:.2f} s')

    @staticmethod
    def generate_hex_geom(hexes_dict,
                          cons_for_lines=[[0, 1], [1, 2], [2, 3], [3, 0], [4, 5], [5, 6], [6, 7], [7, 4], [0, 4], [1, 5], [2, 6], [3, 7]],
                          vol_tag=-1):
        """
        Generates hexahedra volumes from data in hexes dict
        :param hexes_dict: dictionary of data generated by either Winding or FQPL class in Hexahedrons.py
        :param cons_for_lines: line connection defined as list of list in terms of point numbers to connect in the hexahedron
        :param vol_tag: volume tag to use for generated volume of hexahedron
        :return: dictionary with tags of created volumes, surface loops, surface fillings, closed loops, lines, points and channel turn (i.e. turn number of CCT winding).
        """
        tags_dict = {}
        for h_num, h_dict in hexes_dict.hexes.items():
            p_ts = []  # Point tags
            channel_turn = h_dict.pop('ct', None)
            for p_num, p_dict in h_dict.items():
                p_ts.append(gmsh.model.occ.addPoint(p_dict['x'], p_dict['y'], p_dict['z']))

            l_ts = []   # Line tags
            for p_c in cons_for_lines:  # point connections to make a line
                l_ts.append(gmsh.model.occ.addLine(p_ts[p_c[0]], p_ts[p_c[1]]))

            cl_ts = []  # Curved Loops tags
            for l_c in [[0, 1, 2, 3], [4, 5, 6, 7], [3, 8, 7, 11], [1, 9, 5, 10], [0, 8, 4, 9], [2, 11, 6, 10]]:  # line connections to make a curved loop
                cl_ts.append(gmsh.model.occ.addCurveLoop([l_ts[l_c[0]], l_ts[l_c[1]], l_ts[l_c[2]], l_ts[l_c[3]]]))
            sf_ts = []  # Surface Filling tags
            for cl_t in cl_ts:
                sf_ts.append(gmsh.model.occ.addSurfaceFilling(cl_t, degree=3, tol2d=0.000001, tol3d=0.00001, tolAng=0.001, tolCurv=0.05))
            sl_t = gmsh.model.occ.addSurfaceLoop(sf_ts, sewing=True)
            v_t = gmsh.model.occ.addVolume([sl_t], vol_tag)
            tags_dict[int(h_num)] = {'v_t': v_t, 'sl_t': sl_t, 'sf_ts': sf_ts, 'cl_ts': cl_ts, 'l_ts': l_ts, 'p_ts': p_ts, 'ct': channel_turn}
        return tags_dict


class Prepare_BREPs:
    def __init__(self, fdm, verbose=True) -> object:
        """
        Class to preparing brep files by adding terminals.
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """

        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name

        self.verbose = verbose
        self.winding_info_file = os.path.join(self.model_folder, f'{self.magnet_name}.wi')
        self.cctwi = uff.read_data_from_yaml(self.winding_info_file, WindingsInformation)
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()
        self.model_file = os.path.join(self.model_folder, f'{self.magnet_name}.brep')

    @staticmethod
    def get_pts_for_sts(surface_tags):
        """
        Get point tags for surface tags
        :param surface_tags: list of surface tags to get point tags
        :return: list of point tags belonging to the surfaces in the list
        """
        vol_line_tags = []
        vol_line_tags_i = 0
        vol_line_tags_idx = []
        for surf_tag in surface_tags:
            line_tags_new = gmsh.model.getAdjacencies(2, surf_tag)[1]
            for line_tag in line_tags_new:
                vol_line_tags_i += 1
                if line_tag not in vol_line_tags:
                    vol_line_tags.append(line_tag)
                    vol_line_tags_idx.append(vol_line_tags_i)
        point_tags = []
        point_tags_i = 0
        point_tags_idx = []
        for line_tag in vol_line_tags:
            point_tags_new = gmsh.model.getAdjacencies(1, line_tag)[1]
            for point_tag in point_tags_new:
                point_tags_i += 1
                if point_tag not in point_tags:
                    point_tags.append(point_tag)
                    point_tags_idx.append(point_tags_i)
        return point_tags, vol_line_tags_idx, point_tags_idx

    def straighten_terminal(self, gui=False):
        """
        Extends winding geom_generators to the air region boundary by extending the geom_generators with 'terminals' geom_generators up to the air boundary surfaces
        By default this extends along the z axis and makes the final surface normal to the z axis.
        :return: None, saves breps with straighten terminals to disk
        """

        for i, name in enumerate(self.cctwi.windings.names):
            if self.verbose:
                print(f'Straightening terminals of {name} started')
                start_time = timeit.default_timer()
            gmsh.open(os.path.join(self.model_folder, f'{name}.brep'))
            for terminal in [self.cctwi.windings.t_in, self.cctwi.windings.t_out]:
                vol_tag = terminal.vol_st[i]
                vol_surf_tags = gmsh.model.getAdjacencies(3, vol_tag)[1]
                surf_tag = terminal.surf_st[i]
                if surf_tag not in vol_surf_tags:
                    raise ValueError(f'Surface tag of {surf_tag} given for volume in of {vol_tag} does not belong to the volume!')
                # surf_point_tags = self.get_pts_for_sts([surf_tag])      # only 'powered' surface
                surf_point_tags, surf_line_tags_idx, surf_point_tags_idx = self.get_pts_for_sts([surf_tag])  # only 'powered' surface
                vol_point_tags, vol_line_tags_idx, vol_point_tags_idx = self.get_pts_for_sts(vol_surf_tags)  # all tags of hex
                vol_point_dict = {}
                for p_tag in vol_point_tags:
                    xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, p_tag)
                    vol_point_dict[p_tag] = {'x': (xmin + xmax) / 2, 'y': (ymin + ymax) / 2, 'z': (zmin + zmax) / 2}
                z = []
                y = []
                x = []
                for p_tag in surf_point_tags:  # calculate average z position for the terminal surface
                    xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, p_tag)
                    x.append((xmin + xmax) / 2)
                    y.append((ymin + ymax) / 2)
                    z.append((zmin + zmax) / 2)
                x_avg = np.mean(x)
                y_avg = np.mean(y)
                z_avg = np.mean(z)
                sign_x = [math.copysign(1, x_i - x_avg) for x_i in x]
                sign_y = [math.copysign(1, y_i - y_avg) for y_i in y]
                dist_x = []
                dist_y = []
                for j in range(len(x) - 1):
                    dist_x.append(math.sqrt((x[j] - x[j + 1]) ** 2))
                    dist_y.append(math.sqrt((y[j] - y[j + 1]) ** 2))
                eq_len = self.cctdm.geometry.windings.wwhs[i] / 2
                y_shift_sign = math.copysign(1, -terminal.z_air * self.cctdm.geometry.windings.alphas[i])
                new_x = [x_avg + s_x * eq_len for s_x in sign_x]
                new_y = [y_avg + s_y * eq_len + y_shift_sign * eq_len - y_shift_sign * dist_y[0] / 2 for s_y in sign_y]
                for p_tag, x_n, y_n in zip(surf_point_tags, new_x, new_y):  # assign z_avg to only points on the terminal surface
                    vol_point_dict[p_tag]['x'] = x_n
                    vol_point_dict[p_tag]['y'] = y_n
                    vol_point_dict[p_tag]['z'] = z_avg
                gmsh.model.occ.remove([(3, terminal.vol_st[i])], recursive=True)
                gmsh.model.occ.synchronize()
                hexes = Hexahedrons()  # create hex class instance to be able to reuse existing code for generating hexes geom_generators
                hexes.hexes[0] = {}  # only one winding layer is involved per iteration, so hardcoded 0 index
                for p_i, (_, p_dict) in enumerate(vol_point_dict.items()):
                    hexes.hexes[0][p_i] = p_dict
                Generate_BREPs.generate_hex_geom(hexes, cons_for_lines=terminal.lc_st[i], vol_tag=terminal.vol_st[i])
                gmsh.model.occ.synchronize()
            gmsh.write(os.path.join(self.model_folder, f'{name}.brep'))
            if self.verbose:
                print(f'Straightening terminals of {name} took {timeit.default_timer() - start_time:.2f} s')
        # uff.write_data_to_yaml(self.winding_info_yaml_path, self.cctvi.dict())
        if gui:
            gmsh.model.occ.synchronize()
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()

    def extend_terms(self, operation='extend', gui=False, ):
        self.cctwi = uff.read_data_from_yaml(self.winding_info_file, WindingsInformation)  # this is repeated as the operation over the data for the case add, so an original data for extend operation needs to be loaded.
        if operation == 'add':
            for terminal in [self.cctwi.windings.t_in, self.cctwi.windings.t_out]:
                terminal.ndpterms = [1] * len(terminal.ndpterms)
                terminal.z_air = terminal.z_add
        elif operation == 'extend':
            pass
        file_in_postfix = ''
        file_out_postfix = ''
        for i, name in enumerate(self.cctwi.windings.names):
            if self.verbose:
                print(f'Extending terminals of {name} started')
                start_time = timeit.default_timer()
            gmsh.open(os.path.join(self.model_folder, f'{name}{file_in_postfix}.brep'))
            volumes = [vol[1] for vol in gmsh.model.getEntities(dim=3)]
            for oper, terminal in enumerate([self.cctwi.windings.t_in, self.cctwi.windings.t_out]):
                vol_tag = terminal.vol_et[i]
                vol_surf_tags = gmsh.model.getAdjacencies(3, vol_tag)[1]
                surf_tag = terminal.surf_et[i]
                if surf_tag not in vol_surf_tags:
                    raise ValueError(f'Surface tag of {surf_tag} given for volume in of {vol_tag} does not belong to the volume!')
                surf_point_tags, surf_line_tags_idx, surf_point_tags_idx = self.get_pts_for_sts([surf_tag])
                surf_point_tags = sorted(surf_point_tags)
                vol_point_tags, vol_line_tags_idx, vol_point_tags_idx = self.get_pts_for_sts(vol_surf_tags)
                vol_point_tags = sorted(vol_point_tags)
                surf_point_tags_idx_pw = [i for i, e in enumerate(vol_point_tags) if e in set(surf_point_tags)]  # powered
                surf_point_tags_idx_oth = [i for i, e in enumerate(vol_point_tags) if e not in set(surf_point_tags)]  # other (opposite powered)
                vol_point_dict = {}
                for idx_pw, p_tag in zip(surf_point_tags_idx_pw, surf_point_tags):
                    xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, p_tag)
                    vol_point_dict[idx_pw] = {'x': (xmin + xmax) / 2, 'y': (ymin + ymax) / 2, 'z': (zmin + zmax) / 2}
                z_ext = round((terminal.z_air - zmin) / terminal.ndpterms[i], 8)
                for idx_oth, p_tag in zip(surf_point_tags_idx_oth, surf_point_tags):
                    xmin, ymin, zmin, xmax, ymax, zmax = gmsh.model.occ.getBoundingBox(0, p_tag)
                    vol_point_dict[idx_oth] = {'x': (xmin + xmax) / 2, 'y': (ymin + ymax) / 2, 'z': (zmin + zmax) / 2 + z_ext}
                hexes = Hexahedrons()  # create hex class instance to be able to reuse existing code for generating hexes geom_generators
                for h in range(terminal.ndpterms[i]):
                    hexes.hexes[h] = {}
                    for p_i, (_, p_dict) in enumerate(vol_point_dict.items()):
                        hexes.hexes[h][p_i] = copy.deepcopy(p_dict)
                    for p_t, p_dict in vol_point_dict.items():
                        p_dict['z'] = p_dict['z'] + z_ext
                for idx_oth in range(4, 8):
                    z_air = terminal.z_air
                    hexes.hexes[h][idx_oth]['z'] = z_air
                tags_dict = Generate_BREPs.generate_hex_geom(hexes, cons_for_lines=terminal.lc_et[i], vol_tag=-1)
                new_volumes = []
                for _, hex_dict in tags_dict.items():
                    new_volumes.append(hex_dict['v_t'])
                gmsh.model.occ.synchronize()
                if oper == 0:  # this vol_in is needed for reordering volumes later
                    vol_in = new_volumes
            renumbered_vols = []
            other_vols = list(set(volumes) - set(vol_in))
            max_tag = gmsh.model.occ.getMaxTag(3)
            for vol in other_vols:
                gmsh.model.setTag(3, vol, vol + max_tag)
                renumbered_vols.append(vol + max_tag)
            for n_tag, vol in enumerate(reversed(vol_in)):
                gmsh.model.setTag(3, vol, n_tag + 1)
            max_term_tag = len(vol_in) + 1
            spiral_volumes = []
            for n_tag, vol in enumerate(renumbered_vols):
                gmsh.model.setTag(3, vol, n_tag + max_term_tag)
                spiral_volumes.append(n_tag + max_term_tag)
            gmsh.write(os.path.join(self.model_folder, f'{name}{file_out_postfix}.brep'))
            if self.verbose:
                print(f'Straightening terminals of {name} took {timeit.default_timer() - start_time:.2f} s')
                print(f'Writing volume information file: {name}.vi ')
            vi = {'export': spiral_volumes, 'all': [vol[1] for vol in gmsh.model.getEntities(dim=3)]}
            json.dump(vi, open(f"{os.path.join(self.model_folder, name)}.vi", 'w'))
            if self.verbose:
                print(f'Done writing volume information file: {name}.vi ')
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()

    def save_fqpl_vi(self):
        for f_name in self.cctdm.geometry.fqpls.names:
            if self.verbose:
                print(f'Writing volume information file: {f_name}.vi ')
            gmsh.open(os.path.join(self.model_folder, f'{f_name}.brep'))
            volumes = [vol[1] for vol in gmsh.model.getEntities(dim=3)]
            export = volumes
            vi = {'export': export, 'all': volumes}
            json.dump(vi, open(f"{os.path.join(self.model_folder, f_name)}.vi", 'w'))
            if self.verbose:
                print(f'Done writing volume information file: {f_name}.vi ')
        gmsh.clear()

    def fragment(self, gui=False):
        if self.verbose:
            print('Loading files in preparation for fragment operation')
        for f_name in self.cctwi.w_names + self.cctwi.f_names:
            gmsh.merge(os.path.join(self.model_folder, f'{f_name}.brep'))
        num_vol = np.max([vol[1] for vol in gmsh.model.getEntities(dim=3)])
        for f_name in self.cctwi.formers:
            gmsh.merge(os.path.join(self.model_folder, f'{f_name}.brep'))
            num_vol += 1
        gmsh.merge(os.path.join(self.model_folder, f'{self.cctwi.air}.brep'))
        num_vol += 1
        entities = gmsh.model.getEntities(dim=3)
        if len(entities) != num_vol:
            raise ValueError('Not consistent volumes numbers in brep and json files!')
        objectDimTags = [entities[-1]]
        toolDimTags = entities[:-1]
        # central_line = [(1, 3898)]
        # toolDimTags = toolDimTags+central_line
        if self.verbose:
            print(f'Fragmenting {self.magnet_name} started')
            start_time = timeit.default_timer()
        gmsh.model.occ.fragment(objectDimTags, toolDimTags, removeObject=True, removeTool=True)
        gmsh.model.occ.synchronize()
        if self.verbose:
            print(f'Fragmenting {self.magnet_name} took {timeit.default_timer() - start_time:.2f} s')

        gmsh.write(self.model_file)
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def load_geometry(self, gui=False):
        gmsh.open(self.model_file)
        if gui:
            self.gu.launch_interactive_GUI()



import os
import sys
import time
if len(sys.argv) == 3:
    sys.path.insert(0, os.path.join(os.getcwd(), 'steam-fiqus-dev'))
from fiqus.geom_generators.GeometryCCT import Generate_BREPs
from fiqus.geom_generators.GeometryCCT import Prepare_BREPs
from fiqus.pre_processors.PreProcessCCT import Pre_Process
from fiqus.mesh_generators.MeshCCT import Mesh
from fiqus.getdp_runners.RunGetdpCCT import RunGetdpCCT
from fiqus.post_processors.PostProcessCCT import Post_Process
from fiqus.plotters.PlotPythonCCT import PlotPythonCCT

class MainCCT:
    def __init__(self, fdm, verbose=True):
        """
        Main class for working with simulations for CCT type magnets
        :param fdm: FiQuS data model
        :param verbose: if True, more info is printed in the console
        """
        # os.chdir(model_folder)
        self.verbose = verbose
        self.fdm = fdm
        self.settings = None
        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None
        self.model_file = None
        self.model_folder = None

    def generate_geometry(self, gui=False):
        os.chdir(self.geom_folder)
        gb = Generate_BREPs(fdm=self.fdm, verbose=self.verbose)
        gb.generate_windings_or_fqpls('windings')
        gb.save_volume_info()
        gb.generate_windings_or_fqpls('fqpls')
        gb.generate_formers()
        gb.generate_air()
        gb.generate_regions_file()

        pb = Prepare_BREPs(fdm=self.fdm, verbose=self.verbose)
        pb.straighten_terminal(gui=False)
        pb.extend_terms(operation='extend', gui=False)  # use operation='add' for externally generated brep files for windings
        pb.save_fqpl_vi()
        pb.fragment(gui=gui)
        self.model_file = pb.model_file
        self.model_folder = pb.model_folder

    def load_geometry(self, gui=False):
        os.chdir(self.geom_folder)
        pb = Prepare_BREPs(fdm=self.fdm, verbose=self.verbose)
        pb.load_geometry(gui=gui)
        self.model_file = pb.model_file
        self.model_folder = pb.model_folder

    def pre_process(self, gui=False):
        os.chdir(self.geom_folder)
        pp = Pre_Process(fdm=self.fdm)
        pp.calculate_normals(gui=gui)

    def mesh(self, gui=False):
        os.chdir(self.geom_folder)
        pb = Prepare_BREPs(fdm=self.fdm, verbose=self.verbose)
        pb.load_geometry(gui=False)
        os.chdir(self.mesh_folder)
        m = Mesh(fdm=self.fdm)
        m.generate_physical_groups()
        m.generate_mesh(gui=False)
        m.generate_cuts(gui=False)
        m.save_mesh(gui=gui)
        self.model_file = m.model_file
        self.model_folder = m.model_folder
        return {'gamma': 0}  # to be modified with mesh_parameters (see multipole)

    def load_mesh(self, gui=False):
        os.chdir(self.mesh_folder)
        m = Mesh(fdm=self.fdm)
        m.load_mesh(gui=gui)

    def solve_and_postprocess_getdp(self, gui=False):
        os.chdir(self.solution_folder)
        gb = Generate_BREPs(fdm=self.fdm, verbose=self.verbose)
        gb.generate_regions_file()
        rg = RunGetdpCCT(fdm=self.fdm, settings=self.settings)
        rg.assemble_pro()
        start_time = time.time()
        rg.solve_and_postprocess(gui=gui)
        self.model_file = rg.model_file
        self.model_folder = rg.model_folder
        return time.time() - start_time

    def post_process_getdp(self, gui=False):
        os.chdir(self.solution_folder)
        gb = Generate_BREPs(fdm=self.fdm, verbose=self.verbose)
        gb.generate_regions_file()
        rg = RunGetdpCCT(fdm=self.fdm, settings=self.settings)
        rg.assemble_pro()
        rg.postprocess(gui=gui)
        self.model_file = rg.model_file
        self.model_folder = rg.model_folder

    def post_process_python(self, gui=False):
        os.chdir(self.solution_folder)
        pp = Post_Process(self.fdm)
        pp.postprocess_fields(gui=gui)
        #pp.postprocess_thermal_connections()
        pp.postporcess_inductance()
        self.model_file = pp.model_file
        self.model_folder = pp.model_folder
        return {'overall_error': 0}  # to be modified with postprocess_parameters (see multipole)

    def plot_python(self):
        os.chdir(self.solution_folder)
        p = PlotPythonCCT(self.fdm)
        p.plot_elements_file()
